"""Sports reference player model."""

# pylint: disable=too-many-arguments,unused-argument,line-too-long,duplicate-code,too-many-locals,too-many-statements,too-many-branches,broad-exception-caught,too-many-lines
import datetime
import http
import logging
from urllib.parse import unquote

import extruct  # type: ignore
import pytest_is_running
from bs4 import BeautifulSoup
from dateutil.parser import parse
from dateutil.relativedelta import relativedelta
from scrapesession.scrapesession import ScrapeSession  # type: ignore
from scrapesession.session import DEFAULT_TIMEOUT  # type: ignore

from ...cache import MEMORY
from ..google.address_exception import AddressException
from ..google.google_address_model import create_google_address_model
from ..player_model import VERSION, PlayerModel
from ..sex import Sex
from ..species import Species
from .sportsreference_venue_model import create_sportsreference_venue_model

_FIX_URLS = {
    "https://www.sports-reference.com/cbb/players/leyla-öztürk-1.html": "https://www.sports-reference.com/cbb/players/leyla-ozturk-1.html",
    "https://www.sports-reference.com/cbb/players/vianè-cumber-1.html": "https://www.sports-reference.com/cbb/players/viane-cumber-1.html",
    "https://www.sports-reference.com/cbb/players/cia-eklof-1.html": "https://www.sports-reference.com/cbb/players/cia-eklöf-1.html",
    "https://www.sports-reference.com/cbb/players/chae-harris-1.html": "https://www.sports-reference.com/cbb/players/cha%C3%A9-harris-1.html",
    "https://www.sports-reference.com/cbb/players/tilda-sjokvist-1.html": "https://www.sports-reference.com/cbb/players/tilda-sjökvist-1.html",
    "https://www.sports-reference.com/cbb/players/hana-muhl-1.html": "https://www.sports-reference.com/cbb/players/hana-mühl-1.html",
    "https://www.sports-reference.com/cbb/players/noa-comesaña-1.html": "https://www.sports-reference.com/cbb/players/noa-comesana-1.html",
    "https://www.sports-reference.com/cbb/players/nadège-jean-1.html": "https://www.sports-reference.com/cbb/players/nadege-jean-1.html",
}


def _fix_url(url: str) -> str:
    url = unquote(url)
    url = url.replace("é", "e")
    url = url.replace("ć", "c")
    url = url.replace("ã", "a")
    url = url.replace("á", "a")
    url = url.replace("á", "a")
    url = url.replace("ö", "o")
    url = url.replace("ü", "u")

    url = url.replace("Ã©", "é")
    url = url.replace("Ã¶", "ö")
    url = url.replace("Ã¼", "ü")

    return _FIX_URLS.get(url, url)


def _create_sportsreference_player_model(
    session: ScrapeSession,
    player_url: str,
    fg: dict[str, int],
    fga: dict[str, int],
    offensive_rebounds: dict[str, int],
    assists: dict[str, int],
    turnovers: dict[str, int],
    positions: dict[str, str],
    positions_validator: dict[str, str],
    sex: Sex,
    dt: datetime.datetime,
    minutes_played: dict[str, datetime.timedelta],
    three_point_field_goals: dict[str, int],
    three_point_field_goals_attempted: dict[str, int],
    free_throws: dict[str, int],
    free_throws_attempted: dict[str, int],
    defensive_rebounds: dict[str, int],
    steals: dict[str, int],
    blocks: dict[str, int],
    personal_fouls: dict[str, int],
    points: dict[str, int],
    game_scores: dict[str, float],
    point_differentials: dict[str, int],
    version: str,
    goals: dict[str, int],
    penalties_in_minutes: dict[str, datetime.timedelta],
    even_strength_goals: dict[str, int],
    power_play_goals: dict[str, int],
    short_handed_goals: dict[str, int],
    game_winning_goals: dict[str, int],
    even_strength_assists: dict[str, int],
    power_play_assists: dict[str, int],
    short_handed_assists: dict[str, int],
    shots_on_goal: dict[str, int],
    shooting_percentage: dict[str, float],
    shifts: dict[str, int],
    time_on_ice: dict[str, datetime.timedelta],
    decision: dict[str, str],
    goals_against: dict[str, int],
    shots_against: dict[str, int],
    saves: dict[str, int],
    save_percentage: dict[str, float],
    shutouts: dict[str, int],
    individual_corsi_for_events: dict[str, int],
    on_shot_ice_for_events: dict[str, int],
    on_shot_ice_against_events: dict[str, int],
    corsi_for_percentage: dict[str, float],
    relative_corsi_for_percentage: dict[str, float],
    offensive_zone_starts: dict[str, int],
    defensive_zone_starts: dict[str, int],
    offensive_zone_start_percentage: dict[str, float],
    hits: dict[str, int],
    true_shooting_percentage: dict[str, float],
    at_bats: dict[str, int],
    runs_scored: dict[str, int],
    runs_batted_in: dict[str, int],
    bases_on_balls: dict[str, int],
    strikeouts: dict[str, int],
    plate_appearances: dict[str, int],
    hits_at_bats: dict[str, float],
    obp: dict[str, float],
    slg: dict[str, float],
    ops: dict[str, float],
    pitches: dict[str, int],
    strikes: dict[str, int],
    win_probability_added: dict[str, float],
    average_leverage_index: dict[str, float],
    wpa_plus: dict[str, float],
    wpa_minus: dict[str, float],
    cwpa: dict[str, float],
    acli: dict[str, float],
    re24: dict[str, float],
    putouts: dict[str, int],
    innings_pitched: dict[str, int],
    earned_runs: dict[str, int],
    home_runs: dict[str, int],
    era: dict[str, float],
    batters_faced: dict[str, int],
    strikes_by_contact: dict[str, int],
    strikes_swinging: dict[str, int],
    strikes_looking: dict[str, int],
    ground_balls: dict[str, int],
    fly_balls: dict[str, int],
    line_drives: dict[str, int],
    inherited_runners: dict[str, int],
    inherited_scores: dict[str, int],
    effective_field_goal_percentage: dict[str, float],
    penalty_kicks_made: dict[str, int],
    penalty_kicks_attempted: dict[str, int],
    shots_total: dict[str, int],
    shots_on_target: dict[str, int],
    yellow_cards: dict[str, int],
    red_cards: dict[str, int],
    touches: dict[str, int],
    expected_goals: dict[str, float],
    non_penalty_expected_goals: dict[str, float],
    expected_assisted_goals: dict[str, float],
    shot_creating_actions: dict[str, int],
    goal_creating_actions: dict[str, int],
    passes_completed: dict[str, int],
    passes_attempted: dict[str, int],
    pass_completion: dict[str, int],
    progressive_passes: dict[str, int],
    carries: dict[str, int],
    progressive_carries: dict[str, int],
    take_ons_attempted: dict[str, int],
    successful_take_ons: dict[str, int],
    total_passing_distance: dict[str, int],
    progressive_passing_distance: dict[str, int],
    passes_completed_short: dict[str, int],
    passes_attempted_short: dict[str, int],
    pass_completion_short: dict[str, int],
    passes_completed_medium: dict[str, int],
    passes_attempted_medium: dict[str, int],
    pass_completion_medium: dict[str, int],
    passes_completed_long: dict[str, int],
    passes_attempted_long: dict[str, int],
    pass_completion_long: dict[str, int],
    expected_assists: dict[str, float],
    key_passes: dict[str, int],
    passes_into_final_third: dict[str, int],
    passes_into_penalty_area: dict[str, int],
    crosses_into_penalty_area: dict[str, int],
    live_ball_passes: dict[str, int],
    dead_ball_passes: dict[str, int],
    passes_from_free_kicks: dict[str, int],
    through_balls: dict[str, int],
    switches: dict[str, int],
    crosses: dict[str, int],
    throw_ins_taken: dict[str, int],
    corner_kicks: dict[str, int],
    inswinging_corner_kicks: dict[str, int],
    outswinging_corner_kicks: dict[str, int],
    straight_corner_kicks: dict[str, int],
    passes_offside: dict[str, int],
    passes_blocked: dict[str, int],
    tackles_won: dict[str, int],
    tackles_in_defensive_third: dict[str, int],
    tackles_in_middle_third: dict[str, int],
    tackles_in_attacking_third: dict[str, int],
    dribblers_tackled: dict[str, int],
    dribbles_challenged: dict[str, int],
    percent_of_dribblers_tackled: dict[str, float],
    challenges_lost: dict[str, int],
    shots_blocked: dict[str, int],
    tackles_plus_interceptions: dict[str, int],
    errors: dict[str, int],
    touches_in_defensive_penalty_area: dict[str, int],
    touches_in_defensive_third: dict[str, int],
    touches_in_middle_third: dict[str, int],
    touches_in_attacking_third: dict[str, int],
    touches_in_attacking_penalty_area: dict[str, int],
    live_ball_touches: dict[str, int],
    successful_take_on_percentage: dict[str, float],
    times_tackled_during_take_ons: dict[str, int],
    tackled_during_take_ons_percentage: dict[str, int],
    total_carrying_distance: dict[str, int],
    progressive_carrying_distance: dict[str, int],
    carries_into_final_third: dict[str, int],
    carries_into_penalty_area: dict[str, int],
    miscontrols: dict[str, int],
    dispossessed: dict[str, int],
    passes_received: dict[str, int],
    progressive_passes_received: dict[str, int],
    second_yellow_card: dict[str, int],
    fouls_committed: dict[str, int],
    fouls_drawn: dict[str, int],
    offsides: dict[str, int],
    penalty_kicks_won: dict[str, int],
    penalty_kicks_conceded: dict[str, int],
    own_goals: dict[str, int],
    ball_recoveries: dict[str, int],
    aerials_won: dict[str, int],
    aerials_lost: dict[str, int],
    percentage_of_aerials_won: dict[str, float],
    shots_on_target_against: dict[str, int],
    post_shot_expected_goals: dict[str, int],
    passes_attempted_minus_goal_kicks: dict[str, int],
    throws_attempted: dict[str, int],
    percentage_of_passes_that_were_launched: dict[str, float],
    average_pass_length: dict[str, float],
    goal_kicks_attempted: dict[str, int],
    percentage_of_goal_kicks_that_were_launched: dict[str, float],
    average_goal_kick_length: dict[str, float],
    crosses_faced: dict[str, int],
    crosses_stopped: dict[str, int],
    percentage_crosses_stopped: dict[str, float],
    defensive_actions_outside_penalty_area: dict[str, int],
    average_distance_of_defensive_actions: dict[str, float],
    three_point_attempt_rate: dict[str, float],
    tackles: dict[str, int],
    interceptions: dict[str, int],
    clearances: dict[str, int],
    free_throw_attempt_rate: dict[str, float],
    offensive_rebound_percentage: dict[str, float],
    defensive_rebound_percentage: dict[str, float],
    total_rebound_percentage: dict[str, float],
    assist_percentage: dict[str, float],
    steal_percentage: dict[str, float],
    block_percentage: dict[str, float],
    turnover_percentage: dict[str, float],
    usage_percentage: dict[str, float],
    offensive_rating: dict[str, int],
    defensive_rating: dict[str, int],
    box_plus_minus: dict[str, float],
) -> PlayerModel | None:
    """Create a player model from sports reference."""
    player_url = _fix_url(player_url)

    if player_url.endswith("uniform.cgi"):
        return None

    response = session.get(player_url, timeout=DEFAULT_TIMEOUT)
    # Some players can't be accessed on sports reference
    if response.status_code == http.HTTPStatus.FORBIDDEN:
        logging.warning("Cannot access player at URL %s", player_url)
        return None
    response.raise_for_status()
    soup = BeautifulSoup(response.text, "lxml")
    h1 = soup.find("h1")
    if h1 is None:
        logging.warning("h1 is null for %s", player_url)
        return None
    name = h1.get_text().strip()
    data = extruct.extract(response.text, base_url=response.url)
    birth_date = None
    weight = None
    birth_address = None
    height = None
    headshot = None
    for jsonld in data["json-ld"]:
        if jsonld["@type"] != "Person":
            continue
        try:
            birth_date = parse(jsonld["birthDate"])
        except Exception as exc:
            logging.warning(str(exc))
        if "weight" in jsonld:
            weight = float(jsonld["weight"]["value"].split()[0]) * 0.453592
        if "birthPlace" in jsonld:
            try:
                birth_address = create_google_address_model(
                    query=jsonld["birthPlace"],
                    session=session,
                    dt=None,
                )
            except AddressException as exc:
                logging.warning("Failed to find birth address: %s", str(exc))
        if "height" in jsonld:
            height_ft_inches = jsonld["height"]["value"].split()[0].split("-")
            if len(height_ft_inches) == 1:
                height_ft_inches = [
                    x.replace('"', "") for x in height_ft_inches[0].split("'")
                ]
            height = (float(height_ft_inches[0]) * 30.48) + (
                float(height_ft_inches[1]) * 2.54
            )
        if "image" in jsonld:
            headshot = jsonld["image"]["contentUrl"]
    position = positions.get(name)
    seconds_played = None
    if name in minutes_played:
        seconds_played = int(minutes_played[name].total_seconds())
    colleges = {}
    for a in soup.find_all("a"):
        url = a.get("href")
        if url is None:
            continue
        if not url.startswith("/friv/colleges.cgi?college="):
            continue
        title = a.get("title")
        if title in colleges:
            continue
        college = None
        try:
            college = create_sportsreference_venue_model(
                venue_name=title, session=session, dt=dt
            )
        except AddressException as exc:
            logging.warning("Failed to find college: %s", str(exc))
        if college is None:
            continue
        colleges[title] = college
    time_on_ice_dt = time_on_ice.get(name)
    time_on_ice_seconds = time_on_ice_dt.seconds if time_on_ice_dt is not None else None
    penalties_in_minutes_dt = penalties_in_minutes.get(name)
    penalties_in_minutes_seconds = (
        penalties_in_minutes_dt.seconds if penalties_in_minutes_dt is not None else None
    )
    return PlayerModel(
        identifier=name,
        jersey=None,
        kicks=None,
        fumbles=None,
        fumbles_lost=None,
        field_goals=fg.get(name),
        field_goals_attempted=fga.get(name),
        offensive_rebounds=offensive_rebounds.get(name),
        assists=assists.get(name),
        turnovers=turnovers.get(name),
        name=name,
        marks=None,
        handballs=None,
        disposals=None,
        goals=goals.get(name),
        behinds=None,
        hit_outs=None,
        tackles=tackles.get(name),
        rebounds=None,
        insides=None,
        clearances=clearances.get(name),
        clangers=None,
        free_kicks_for=None,
        free_kicks_against=None,
        brownlow_votes=None,
        contested_possessions=None,
        uncontested_possessions=None,
        contested_marks=None,
        marks_inside=None,
        one_percenters=None,
        bounces=None,
        goal_assists=None,
        percentage_played=None,
        birth_date=birth_date,
        species=str(Species.HUMAN),
        handicap_weight=None,
        father=None,
        sex=str(sex),
        age=None if birth_date is None else relativedelta(birth_date, dt).years,
        starting_position=positions_validator[position]
        if position is not None
        else None,
        weight=weight,
        birth_address=birth_address,
        owner=None,
        seconds_played=seconds_played,
        three_point_field_goals=three_point_field_goals.get(name),
        three_point_field_goals_attempted=three_point_field_goals_attempted.get(name),
        free_throws=free_throws.get(name),
        free_throws_attempted=free_throws_attempted.get(name),
        defensive_rebounds=defensive_rebounds.get(name),
        steals=steals.get(name),
        blocks=blocks.get(name),
        personal_fouls=personal_fouls.get(name),
        points=points.get(name),
        game_score=game_scores.get(name),
        point_differential=point_differentials.get(name),
        version=version,
        height=height,
        colleges=list(colleges.values()),
        headshot=headshot,
        forced_fumbles=None,
        fumbles_recovered=None,
        fumbles_recovered_yards=None,
        fumbles_touchdowns=None,
        offensive_two_point_returns=None,
        offensive_fumbles_touchdowns=None,
        defensive_fumbles_touchdowns=None,
        average_gain=None,
        completion_percentage=None,
        completions=None,
        espn_quarterback_rating=None,
        interception_percentage=None,
        interceptions=interceptions.get(name),
        long_passing=None,
        misc_yards=None,
        net_passing_yards=None,
        net_total_yards=None,
        passing_attempts=None,
        passing_big_plays=None,
        passing_first_downs=None,
        passing_fumbles=None,
        passing_fumbles_lost=None,
        passing_touchdown_percentage=None,
        passing_touchdowns=None,
        passing_yards=None,
        passing_yards_after_catch=None,
        quarterback_rating=None,
        sacks=None,
        passing_yards_at_catch=None,
        sacks_yards_lost=None,
        net_passing_attempts=None,
        total_offensive_plays=None,
        total_points=None,
        total_touchdowns=None,
        total_yards=None,
        total_yards_from_scrimmage=None,
        two_point_pass=None,
        two_point_pass_attempt=None,
        yards_per_completion=None,
        yards_per_pass_attempt=None,
        net_yards_per_pass_attempt=None,
        long_rushing=None,
        rushing_attempts=None,
        rushing_big_plays=None,
        rushing_first_downs=None,
        rushing_fumbles=None,
        rushing_fumbles_lost=None,
        rushing_touchdowns=None,
        rushing_yards=None,
        stuffs=None,
        stuff_yards_lost=None,
        two_point_rush=None,
        two_point_rush_attempts=None,
        yards_per_rush_attempt=None,
        espn_widereceiver=None,
        long_reception=None,
        receiving_big_plays=None,
        receiving_first_downs=None,
        receiving_fumbles=None,
        receiving_fumbles_lost=None,
        receiving_targets=None,
        receiving_touchdowns=None,
        receiving_yards=None,
        receiving_yards_after_catch=None,
        receiving_yards_at_catch=None,
        receptions=None,
        two_point_receptions=None,
        two_point_reception_attempts=None,
        yards_per_reception=None,
        assist_tackles=None,
        average_interception_yards=None,
        average_sack_yards=None,
        average_stuff_yards=None,
        blocked_field_goal_touchdowns=None,
        blocked_punt_touchdowns=None,
        defensive_touchdowns=None,
        hurries=None,
        kicks_blocked=None,
        long_interception=None,
        misc_touchdowns=None,
        passes_batted_down=None,
        passes_defended=None,
        quarterback_hits=None,
        sacks_assisted=None,
        sacks_unassisted=None,
        sacks_yards=None,
        safeties=None,
        solo_tackles=None,
        stuff_yards=None,
        tackles_for_loss=None,
        tackles_yards_lost=None,
        yards_allowed=None,
        points_allowed=None,
        one_point_safeties_made=None,
        missed_field_goal_return_td=None,
        blocked_punt_ez_rec_td=None,
        interception_touchdowns=None,
        interception_yards=None,
        average_kickoff_return_yards=None,
        average_kickoff_yards=None,
        extra_point_attempts=None,
        extra_point_percentage=None,
        extra_point_blocked=None,
        extra_points_blocked_percentage=None,
        extra_points_made=None,
        fair_catches=None,
        fair_catch_percentage=None,
        field_goal_attempts_max_19_yards=None,
        field_goal_attempts_max_29_yards=None,
        field_goal_attempts_max_39_yards=None,
        field_goal_attempts_max_49_yards=None,
        field_goal_attempts_max_59_yards=None,
        field_goal_attempts_max_99_yards=None,
        field_goal_attempts_above_50_yards=None,
        field_goal_attempt_yards=None,
        field_goals_blocked=None,
        field_goals_blocked_percentage=None,
        field_goals_made=None,
        field_goals_made_max_19_yards=None,
        field_goals_made_max_29_yards=None,
        field_goals_made_max_39_yards=None,
        field_goals_made_max_49_yards=None,
        field_goals_made_max_59_yards=None,
        field_goals_made_max_99_yards=None,
        field_goals_made_above_50_yards=None,
        field_goals_made_yards=None,
        field_goals_missed_yards=None,
        kickoff_out_of_bounds=None,
        kickoff_returns=None,
        kickoff_returns_touchdowns=None,
        kickoff_return_yards=None,
        kickoffs=None,
        kickoff_yards=None,
        long_field_goal_attempt=None,
        long_field_goal_made=None,
        long_kickoff=None,
        total_kicking_points=None,
        touchback_percentage=None,
        touchbacks=None,
        defensive_fumble_returns=None,
        defensive_fumble_return_yards=None,
        fumble_recoveries=None,
        fumble_recovery_yards=None,
        kick_return_fair_catches=None,
        kick_return_fair_catch_percentage=None,
        kick_return_fumbles=None,
        kick_return_fumbles_lost=None,
        kick_returns=None,
        kick_return_touchdowns=None,
        kick_return_yards=None,
        long_kick_return=None,
        long_punt_return=None,
        misc_fumble_returns=None,
        misc_fumble_return_yards=None,
        opposition_fumble_recoveries=None,
        opposition_fumble_recovery_yards=None,
        opposition_special_team_fumble_returns=None,
        opposition_special_team_fumble_return_yards=None,
        punt_return_fair_catches=None,
        punt_return_fair_catch_percentage=None,
        punt_return_fumbles=None,
        punt_return_fumbles_lost=None,
        punt_returns=None,
        punt_returns_started_inside_the_10=None,
        punt_returns_started_inside_the_20=None,
        punt_return_touchdowns=None,
        special_team_fumble_returns=None,
        yards_per_kick_return=None,
        yards_per_punt_return=None,
        yards_per_return=None,
        average_punt_return_yards=None,
        gross_average_punt_yards=None,
        long_punt=None,
        net_average_punt_yards=None,
        punts=None,
        punts_blocked=None,
        punts_blocked_percentage=None,
        punts_inside_10=None,
        punts_inside_10_percentage=None,
        punts_inside_20=None,
        punts_inside_20_percentage=None,
        punts_over_50=None,
        punt_yards=None,
        defensive_points=None,
        misc_points=None,
        return_touchdowns=None,
        total_two_point_conversions=None,
        passing_touchdowns_9_yards=None,
        passing_touchdowns_19_yards=None,
        passing_touchdowns_29_yards=None,
        passing_touchdowns_39_yards=None,
        passing_touchdowns_49_yards=None,
        passing_touchdowns_above_50_yards=None,
        receiving_touchdowns_9_yards=None,
        receiving_touchdowns_19_yards=None,
        receiving_touchdowns_29_yards=None,
        receiving_touchdowns_39_yards=None,
        punt_return_yards=None,
        receiving_touchdowns_49_yards=None,
        receiving_touchdowns_above_50_yards=None,
        rushing_touchdowns_9_yards=None,
        rushing_touchdowns_19_yards=None,
        rushing_touchdowns_29_yards=None,
        rushing_touchdowns_39_yards=None,
        rushing_touchdowns_49_yards=None,
        rushing_touchdowns_above_50_yards=None,
        penalties_in_minutes=penalties_in_minutes_seconds,
        even_strength_goals=even_strength_goals.get(name),
        power_play_goals=power_play_goals.get(name),
        short_handed_goals=short_handed_goals.get(name),
        game_winning_goals=game_winning_goals.get(name),
        even_strength_assists=even_strength_assists.get(name),
        power_play_assists=power_play_assists.get(name),
        short_handed_assists=short_handed_assists.get(name),
        shots_on_goal=shots_on_goal.get(name),
        shooting_percentage=shooting_percentage.get(name),
        shifts=shifts.get(name),
        time_on_ice=time_on_ice_seconds,
        decision=decision.get(name),
        goals_against=goals_against.get(name),
        shots_against=shots_against.get(name),
        saves=saves.get(name),
        save_percentage=save_percentage.get(name),
        shutouts=shutouts.get(name),
        individual_corsi_for_events=individual_corsi_for_events.get(name),
        on_shot_ice_for_events=on_shot_ice_for_events.get(name),
        on_shot_ice_against_events=on_shot_ice_against_events.get(name),
        corsi_for_percentage=corsi_for_percentage.get(name),
        relative_corsi_for_percentage=relative_corsi_for_percentage.get(name),
        offensive_zone_starts=offensive_zone_starts.get(name),
        defensive_zone_starts=defensive_zone_starts.get(name),
        offensive_zone_start_percentage=offensive_zone_start_percentage.get(name),
        hits=hits.get(name),
        true_shooting_percentage=true_shooting_percentage.get(name),
        at_bats=at_bats.get(name),
        runs_scored=runs_scored.get(name),
        runs_batted_in=runs_batted_in.get(name),
        bases_on_balls=bases_on_balls.get(name),
        strikeouts=strikeouts.get(name),
        plate_appearances=plate_appearances.get(name),
        hits_at_bats=hits_at_bats.get(name),
        obp=obp.get(name),
        slg=slg.get(name),
        ops=ops.get(name),
        pitches=pitches.get(name),
        strikes=strikes.get(name),
        win_probability_added=win_probability_added.get(name),
        average_leverage_index=average_leverage_index.get(name),
        wpa_plus=wpa_plus.get(name),
        wpa_minus=wpa_minus.get(name),
        cwpa=cwpa.get(name),
        acli=acli.get(name),
        re24=re24.get(name),
        putouts=putouts.get(name),
        innings_pitched=innings_pitched.get(name),
        earned_runs=earned_runs.get(name),
        home_runs=home_runs.get(name),
        era=era.get(name),
        batters_faced=batters_faced.get(name),
        strikes_by_contact=strikes_by_contact.get(name),
        strikes_swinging=strikes_swinging.get(name),
        strikes_looking=strikes_looking.get(name),
        ground_balls=ground_balls.get(name),
        fly_balls=fly_balls.get(name),
        line_drives=line_drives.get(name),
        inherited_runners=inherited_runners.get(name),
        inherited_scores=inherited_scores.get(name),
        effective_field_goal_percentage=effective_field_goal_percentage.get(name),
        penalty_kicks_made=penalty_kicks_made.get(name),
        penalty_kicks_attempted=penalty_kicks_attempted.get(name),
        shots_total=shots_total.get(name),
        shots_on_target=shots_on_target.get(name),
        yellow_cards=yellow_cards.get(name),
        red_cards=red_cards.get(name),
        touches=touches.get(name),
        expected_goals=expected_goals.get(name),
        non_penalty_expected_goals=non_penalty_expected_goals.get(name),
        expected_assisted_goals=expected_assisted_goals.get(name),
        shot_creating_actions=shot_creating_actions.get(name),
        goal_creating_actions=goal_creating_actions.get(name),
        passes_completed=passes_completed.get(name),
        passes_attempted=passes_attempted.get(name),
        pass_completion=pass_completion.get(name),
        progressive_passes=progressive_passes.get(name),
        carries=carries.get(name),
        progressive_carries=progressive_carries.get(name),
        take_ons_attempted=take_ons_attempted.get(name),
        successful_take_ons=successful_take_ons.get(name),
        total_passing_distance=total_passing_distance.get(name),
        progressive_passing_distance=progressive_passing_distance.get(name),
        passes_completed_short=passes_completed_short.get(name),
        passes_attempted_short=passes_attempted_short.get(name),
        pass_completion_short=pass_completion_short.get(name),
        passes_completed_medium=passes_completed_medium.get(name),
        passes_attempted_medium=passes_attempted_medium.get(name),
        pass_completion_medium=pass_completion_medium.get(name),
        passes_completed_long=passes_completed_long.get(name),
        passes_attempted_long=passes_attempted_long.get(name),
        pass_completion_long=pass_completion_long.get(name),
        expected_assists=expected_assists.get(name),
        key_passes=key_passes.get(name),
        passes_into_final_third=passes_into_final_third.get(name),
        passes_into_penalty_area=passes_into_penalty_area.get(name),
        crosses_into_penalty_area=crosses_into_penalty_area.get(name),
        live_ball_passes=live_ball_passes.get(name),
        dead_ball_passes=dead_ball_passes.get(name),
        passes_from_free_kicks=passes_from_free_kicks.get(name),
        through_balls=through_balls.get(name),
        switches=switches.get(name),
        crosses=crosses.get(name),
        throw_ins_taken=throw_ins_taken.get(name),
        corner_kicks=corner_kicks.get(name),
        inswinging_corner_kicks=inswinging_corner_kicks.get(name),
        outswinging_corner_kicks=outswinging_corner_kicks.get(name),
        straight_corner_kicks=straight_corner_kicks.get(name),
        passes_offside=passes_offside.get(name),
        passes_blocked=passes_blocked.get(name),
        tackles_won=tackles_won.get(name),
        tackles_in_defensive_third=tackles_in_defensive_third.get(name),
        tackles_in_middle_third=tackles_in_middle_third.get(name),
        tackles_in_attacking_third=tackles_in_attacking_third.get(name),
        dribblers_tackled=dribblers_tackled.get(name),
        dribbles_challenged=dribbles_challenged.get(name),
        percent_of_dribblers_tackled=percent_of_dribblers_tackled.get(name),
        challenges_lost=challenges_lost.get(name),
        shots_blocked=shots_blocked.get(name),
        tackles_plus_interceptions=tackles_plus_interceptions.get(name),
        errors=errors.get(name),
        touches_in_defensive_penalty_area=touches_in_defensive_penalty_area.get(name),
        touches_in_defensive_third=touches_in_defensive_third.get(name),
        touches_in_middle_third=touches_in_middle_third.get(name),
        touches_in_attacking_third=touches_in_attacking_third.get(name),
        touches_in_attacking_penalty_area=touches_in_attacking_penalty_area.get(name),
        live_ball_touches=live_ball_touches.get(name),
        successful_take_on_percentage=successful_take_on_percentage.get(name),
        times_tackled_during_take_ons=times_tackled_during_take_ons.get(name),
        tackled_during_take_on_percentage=tackled_during_take_ons_percentage.get(name),
        total_carrying_distance=total_carrying_distance.get(name),
        progressive_carrying_distance=progressive_carrying_distance.get(name),
        carries_into_final_third=carries_into_final_third.get(name),
        carries_into_penalty_area=carries_into_penalty_area.get(name),
        miscontrols=miscontrols.get(name),
        dispossessed=dispossessed.get(name),
        passes_received=passes_received.get(name),
        progressive_passes_received=progressive_passes_received.get(name),
        second_yellow_card=second_yellow_card.get(name),
        fouls_committed=fouls_committed.get(name),
        fouls_drawn=fouls_drawn.get(name),
        offsides=offsides.get(name),
        penalty_kicks_won=penalty_kicks_won.get(name),
        penalty_kicks_conceded=penalty_kicks_conceded.get(name),
        own_goals=own_goals.get(name),
        ball_recoveries=ball_recoveries.get(name),
        aerials_won=aerials_won.get(name),
        aerials_lost=aerials_lost.get(name),
        percentage_of_aerials_won=percentage_of_aerials_won.get(name),
        shots_on_target_against=shots_on_target_against.get(name),
        post_shot_expected_goals=post_shot_expected_goals.get(name),
        passes_attempted_minus_goal_kicks=passes_attempted_minus_goal_kicks.get(name),
        throws_attempted=throws_attempted.get(name),
        percentage_of_passes_that_were_launched=percentage_of_passes_that_were_launched.get(
            name
        ),
        average_pass_length=average_pass_length.get(name),
        goal_kicks_attempted=goal_kicks_attempted.get(name),
        percentage_of_goal_kicks_that_were_launched=percentage_of_goal_kicks_that_were_launched.get(
            name
        ),
        average_goal_kick_length=average_goal_kick_length.get(name),
        crosses_faced=crosses_faced.get(name),
        crosses_stopped=crosses_stopped.get(name),
        percentage_crosses_stopped=percentage_crosses_stopped.get(name),
        defensive_actions_outside_penalty_area=defensive_actions_outside_penalty_area.get(
            name
        ),
        average_distance_of_defensive_actions=average_distance_of_defensive_actions.get(
            name
        ),
        three_point_attempt_rate=three_point_attempt_rate.get(name),
        batting_style=None,
        bowling_style=None,
        playing_roles=None,
        runs=None,
        balls=None,
        fours=None,
        sixes=None,
        strikerate=None,
        fall_of_wicket_order=None,
        fall_of_wicket_num=None,
        fall_of_wicket_runs=None,
        fall_of_wicket_balls=None,
        fall_of_wicket_overs=None,
        fall_of_wicket_over_number=None,
        ball_over_actual=None,
        ball_over_unique=None,
        ball_total_runs=None,
        ball_batsman_runs=None,
        overs=None,
        maidens=None,
        conceded=None,
        wickets=None,
        economy=None,
        runs_per_ball=None,
        dots=None,
        wides=None,
        no_balls=None,
        free_throw_attempt_rate=free_throw_attempt_rate.get(name),
        offensive_rebound_percentage=offensive_rebound_percentage.get(name),
        defensive_rebound_percentage=defensive_rebound_percentage.get(name),
        total_rebound_percentage=total_rebound_percentage.get(name),
        assist_percentage=assist_percentage.get(name),
        steal_percentage=steal_percentage.get(name),
        block_percentage=block_percentage.get(name),
        turnover_percentage=turnover_percentage.get(name),
        usage_percentage=usage_percentage.get(name),
        offensive_rating=offensive_rating.get(name),
        defensive_rating=defensive_rating.get(name),
        box_plus_minus=box_plus_minus.get(name),
        ace_percentage=None,
        double_fault_percentage=None,
        first_serves_in=None,
        first_serve_percentage=None,
        second_serve_percentage=None,
        break_points_saved=None,
        return_points_won_percentage=None,
        winners=None,
        winners_fronthand=None,
        winners_backhand=None,
        unforced_errors=None,
        unforced_errors_fronthand=None,
        unforced_errors_backhand=None,
        serve_points=None,
        serves_won=None,
        serves_aces=None,
        serves_unreturned=None,
        serves_forced_error_percentage=None,
        serves_won_in_three_shots_or_less=None,
        serves_wide_percentage=None,
        serves_body_percentage=None,
        serves_t_percentage=None,
        serves_wide_deuce_percentage=None,
        serves_body_deuce_percentage=None,
        serves_t_deuce_percentage=None,
        serves_wide_ad_percentage=None,
        serves_body_ad_percentage=None,
        serves_t_ad_percentage=None,
        serves_net_percentage=None,
        serves_wide_direction_percentage=None,
        shots_deep_percentage=None,
        shots_deep_wide_percentage=None,
        shots_foot_errors_percentage=None,
        shots_unknown_percentage=None,
        points_won_percentage=None,
        goal_accuracy=None,
        tackles_inside_50=None,
        total_possessions=None,
        score_involvements=None,
        centre_clearances=None,
        disposal_efficiency=None,
        stoppage_clearances=None,
        uncontested_marks=None,
        accurate_crosses=None,
        accurate_long_balls=None,
        accurate_passes=None,
        accurate_through_balls=None,
        cross_percentage=None,
        free_kick_goals=None,
        free_kick_percentage=None,
        free_kick_shots=None,
        game_winning_assists=None,
        headed_goals=None,
        inaccurate_crosses=None,
        inaccurate_long_balls=None,
        inaccurate_passes=None,
        inaccurate_through_balls=None,
        left_footed_shots=None,
        long_ball_percentage=None,
        penalty_kick_goals=None,
        penalty_kick_percentage=None,
        penalty_kicks_missed=None,
        possession_percentage=None,
        possession_time=None,
        right_footed_shots=None,
        shoot_out_goals=None,
        shoot_out_misses=None,
        shoot_out_percentage=None,
        shot_assists=None,
        shot_percentage=None,
        shots_headed=None,
        shots_off_target=None,
        shots_on_post=None,
        through_ball_percentage=None,
        long_balls=None,
        total_passes=None,
        average_rating_from_editor=None,
        average_rating_from_user=None,
        did_not_play=None,
        draws=None,
        goal_difference=None,
        losses=None,
        lost_corners=None,
        minutes=None,
        pass_percentage=None,
        starts=None,
        sub_ins=None,
        sub_outs=None,
        suspensions=None,
        time_ended=None,
        time_started=None,
        win_percentage=None,
        wins=None,
        won_corners=None,
        clean_sheet=None,
        crosses_caught=None,
        goals_conceded=None,
        partial_clean_sheet=None,
        penalty_kick_conceded=None,
        penalty_kick_save_percentage=None,
        penalty_kicks_faced=None,
        penalty_kicks_saved=None,
        punches=None,
        shoot_out_kicks_faced=None,
        shoot_out_kicks_saved=None,
        shoot_out_save_percentage=None,
        shots_faced=None,
        smothers=None,
        unclaimed_crosses=None,
        effective_clearances=None,
        effective_tackles=None,
        ineffective_tackles=None,
        tackle_percentage=None,
        appearances=None,
        average_rating_from_correspondent=None,
        average_rating_from_data_feed=None,
        games_played=None,
        team_games_played=None,
        hit_by_pitch=None,
        rbis=None,
        sac_hits=None,
        stolen_bases=None,
        walks=None,
        catcher_interference=None,
        gidps=None,
        sac_flies=None,
        grand_slam_home_runs=None,
        runners_left_on_base=None,
        triples=None,
        game_winning_rbis=None,
        intentional_walks=None,
        doubles=None,
        caught_stealing=None,
        games_started=None,
        pinch_at_bats=None,
        pinch_hits=None,
        player_rating=None,
        is_qualified=None,
        is_qualified_steals=None,
        total_bases=None,
        projected_home_runs=None,
        extra_base_hits=None,
        runs_created=None,
        batting_average=None,
        pinch_average=None,
        slug_average=None,
        secondary_average=None,
        on_base_percentage=None,
        ground_to_fly_ratio=None,
        runs_created_per_27_outs=None,
        batter_rating=None,
        at_bats_per_home_run=None,
        stolen_base_percentage=None,
        pitches_per_plate_appearance=None,
        isolated_power=None,
        walk_to_strikeout_ratio=None,
        walks_per_plate_appearance=None,
        secondary_average_minus_batting_average=None,
        runs_produced=None,
        runs_ratio=None,
        patience_ratio=None,
        balls_in_play_average=None,
        mlb_rating=None,
        offensive_wins_above_replacement=None,
        wins_above_replacement=None,
        batters_hit=None,
        sacrifice_bunts=None,
        save_opportunities=None,
        finishes=None,
        balks=None,
        holds=None,
        complete_games=None,
        perfect_games=None,
        wild_pitches=None,
        third_innings=None,
        team_earned_runs=None,
        pickoff_attempts=None,
        run_support=None,
        pitches_as_starter=None,
        average_game_score=None,
        quality_starts=None,
        inherited_runners_scored=None,
        opponent_total_bases=None,
        is_qualified_saves=None,
        full_innings=None,
        part_innings=None,
        blown_saves=None,
        innings=None,
        whip=None,
        caught_stealing_percentage=None,
        pitches_per_start=None,
        pitches_per_inning=None,
        run_support_average=None,
        opponent_average=None,
        opponent_slug_average=None,
        opponent_on_base_percentage=None,
        opponent_ops=None,
        strikeouts_per_nine_innings=None,
        strikeout_to_walk_ratio=None,
        tough_losses=None,
        cheap_wins=None,
        save_opportunities_per_win=None,
        pitch_count=None,
        strike_pitch_ratio=None,
        double_plays=None,
        opportunities=None,
        passed_balls=None,
        outfield_assists=None,
        pickoffs=None,
        outs_on_field=None,
        triple_plays=None,
        balls_in_zone=None,
        extra_bases=None,
        outs_made=None,
        catcher_third_innings_played=None,
        catcher_caught_stealing=None,
        catcher_stolen_bases_allowed=None,
        catcher_earned_runs=None,
        is_qualified_catcher=None,
        is_qualified_pitcher=None,
        successful_chances=None,
        total_chances=None,
        full_innings_played=None,
        part_innings_played=None,
        fielding_percentage=None,
        range_factor=None,
        zone_rating=None,
        catcher_caught_stealing_percentage=None,
        catcher_era=None,
        def_warbr=None,
        #
        average_three_point_field_goals_attempted=None,
        average_free_throws_made=None,
        average_free_throws_attempted=None,
        average_points=None,
        average_offensive_rebounds=None,
        average_assists=None,
        average_turnovers=None,
        estimated_possessions=None,
        average_estimated_possessions=None,
        points_per_estimated_possessions=None,
        average_team_turnovers=None,
        average_total_turnovers=None,
        three_point_field_goal_percentage=None,
        two_point_field_goals_made=None,
        two_point_field_goals_attempted=None,
        average_two_point_field_goals_made=None,
        average_two_point_field_goals_attempted=None,
        two_point_field_goal_percentage=None,
        shooting_efficiency=None,
        scoring_efficiency=None,
        average_48_field_goals_made=None,
        average_48_field_goals_attempted=None,
        average_48_three_point_field_goals_made=None,
        average_48_three_point_field_goals_attempted=None,
        average_48_free_throws_made=None,
        average_48_free_throws_attempted=None,
        average_48_points=None,
        average_48_offensive_rebounds=None,
        average_48_assists=None,
        average_48_turnovers=None,
        p40=None,
        a40=None,
        average_rebounds=None,
        average_fouls=None,
        average_flagrant_fouls=None,
        average_technical_fouls=None,
        average_ejections=None,
        average_disqualifications=None,
        assist_turnover_ratio=None,
        steal_foul_ratio=None,
        block_foul_ratio=None,
        average_team_rebounds=None,
        total_technical_fouls=None,
        team_assist_turnover_ratio=None,
        steal_turnover_ratio=None,
        average_48_rebounds=None,
        average_48_fouls=None,
        average_48_flagrant_fouls=None,
        average_48_technical_fouls=None,
        average_48_ejections=None,
        average_48_disqualifications=None,
        r40=None,
        double_double=None,
        triple_double=None,
        free_throws_made=None,
        three_point_percentage=None,
        three_point_field_goals_made=None,
        total_turnovers=None,
        points_in_paint=None,
        brick_index=None,
        average_field_goals_made=None,
        average_field_goals_attempted=None,
        average_three_point_field_goals_made=None,
        average_defensive_rebounds=None,
        average_blocks=None,
        average_steals=None,
        average_48_defensive_rebounds=None,
        average_48_blocks=None,
        average_48_steals=None,
        largest_lead=None,
        disqualifications=None,
        flagrant_fouls=None,
        fouls=None,
        ejections=None,
        technical_fouls=None,
        average_minutes=None,
        nba_rating=None,
        plus_minus=None,
        even_strength_time_on_ice=None,
        shifts_per_game=None,
        production=None,
        shot_differential=None,
        goal_differential=None,
        pim_differential=None,
        rating=None,
        average_goals=None,
        ytd_goals=None,
        shots_in_first_period=None,
        shots_in_second_period=None,
        shots_in_third_period=None,
        shots_overtime=None,
        average_shots=None,
        points_per_game=None,
        power_play_opportunities=None,
        power_play_percentage=None,
        shootout_attempts=None,
        shootout_shot_percentage=None,
        empty_net_goals_for=None,
        shutouts_against=None,
        total_face_offs=None,
        average_goals_against=None,
        average_shots_against=None,
        penalty_kill_percentage=None,
        power_play_goals_against=None,
        short_handed_goals_against=None,
        shootout_saves=None,
        shootout_shots_against=None,
        times_short_handed=None,
        empty_net_goals_against=None,
        overtime_losses=None,
        shots_missed=None,
        minor_penalties=None,
        match_penalties=None,
        misconducts=None,
        game_misconducts=None,
        boarding_penalties=None,
        unsportsmanlike_penalties=None,
        fighting_penalties=None,
        average_fights=None,
        time_between_fights=None,
        instigator_penalties=None,
        charging_penalties=None,
        hooking_penalties=None,
        tripping_penalties=None,
        roughing_penalties=None,
        holding_penalties=None,
        interference_penalties=None,
        slashing_penalties=None,
        high_sticking_penalties=None,
        cross_checking_penalties=None,
        stick_holding_penalties=None,
        goalie_interference_penalties=None,
        elbowing_penalties=None,
        diving_penalties=None,
        takeaways=None,
        even_strength_saves=None,
        power_play_saves=None,
        short_handed_saves=None,
        games=None,
        game_started=None,
        ties=None,
        time_on_ice_per_game=None,
        power_play_time_on_ice=None,
        short_handed_time_on_ice=None,
        faceoffs_won=None,
        faceoffs_lost=None,
        faceoff_percentage=None,
        unassisted_goals=None,
        game_tying_goals=None,
        giveaways=None,
        penalties=None,
        penalty_minutes=None,
        penalty_minutes_against=None,
        major_penalties=None,
        net_passing_yards_per_game=None,
        net_yards_per_game=None,
        passing_yards_per_game=None,
        total_points_per_game=None,
        yards_from_scrimmage_per_game=None,
        yards_per_game=None,
        espn_rb_rating=None,
        rushing_yards_per_game=None,
        receiving_yards_per_game=None,
        two_point_returns=None,
        field_goal_attempts=None,
        special_team_fumble_return_yards=None,
        kick_extra_points=None,
        kick_extra_points_made=None,
        attempts_in_box=None,
        second_assists=None,
        qbr=None,
        attempts_out_box=None,
        adjusted_qbr=None,
        turnover_points=None,
        fantasy_rating=None,
        team_turnovers=None,
        second_chance_points=None,
        fast_break_points=None,
        team_rebounds=None,
        gained=None,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_sportsreference_player_model(
    session: ScrapeSession,
    player_url: str,
    fg: dict[str, int],
    fga: dict[str, int],
    offensive_rebounds: dict[str, int],
    assists: dict[str, int],
    turnovers: dict[str, int],
    positions: dict[str, str],
    positions_validator: dict[str, str],
    sex: Sex,
    dt: datetime.datetime,
    minutes_played: dict[str, datetime.timedelta],
    three_point_field_goals: dict[str, int],
    three_point_field_goals_attempted: dict[str, int],
    free_throws: dict[str, int],
    free_throws_attempted: dict[str, int],
    defensive_rebounds: dict[str, int],
    steals: dict[str, int],
    blocks: dict[str, int],
    personal_fouls: dict[str, int],
    points: dict[str, int],
    game_scores: dict[str, float],
    point_differentials: dict[str, int],
    version: str,
    goals: dict[str, int],
    penalties_in_minutes: dict[str, datetime.timedelta],
    even_strength_goals: dict[str, int],
    power_play_goals: dict[str, int],
    short_handed_goals: dict[str, int],
    game_winning_goals: dict[str, int],
    even_strength_assists: dict[str, int],
    power_play_assists: dict[str, int],
    short_handed_assists: dict[str, int],
    shots_on_goal: dict[str, int],
    shooting_percentage: dict[str, float],
    shifts: dict[str, int],
    time_on_ice: dict[str, datetime.timedelta],
    decision: dict[str, str],
    goals_against: dict[str, int],
    shots_against: dict[str, int],
    saves: dict[str, int],
    save_percentage: dict[str, float],
    shutouts: dict[str, int],
    individual_corsi_for_events: dict[str, int],
    on_shot_ice_for_events: dict[str, int],
    on_shot_ice_against_events: dict[str, int],
    corsi_for_percentage: dict[str, float],
    relative_corsi_for_percentage: dict[str, float],
    offensive_zone_starts: dict[str, int],
    defensive_zone_starts: dict[str, int],
    offensive_zone_start_percentage: dict[str, float],
    hits: dict[str, int],
    true_shooting_percentage: dict[str, float],
    at_bats: dict[str, int],
    runs_scored: dict[str, int],
    runs_batted_in: dict[str, int],
    bases_on_balls: dict[str, int],
    strikeouts: dict[str, int],
    plate_appearances: dict[str, int],
    hits_at_bats: dict[str, float],
    obp: dict[str, float],
    slg: dict[str, float],
    ops: dict[str, float],
    pitches: dict[str, int],
    strikes: dict[str, int],
    win_probability_added: dict[str, float],
    average_leverage_index: dict[str, float],
    wpa_plus: dict[str, float],
    wpa_minus: dict[str, float],
    cwpa: dict[str, float],
    acli: dict[str, float],
    re24: dict[str, float],
    putouts: dict[str, int],
    innings_pitched: dict[str, int],
    earned_runs: dict[str, int],
    home_runs: dict[str, int],
    era: dict[str, float],
    batters_faced: dict[str, int],
    strikes_by_contact: dict[str, int],
    strikes_swinging: dict[str, int],
    strikes_looking: dict[str, int],
    ground_balls: dict[str, int],
    fly_balls: dict[str, int],
    line_drives: dict[str, int],
    inherited_runners: dict[str, int],
    inherited_scores: dict[str, int],
    effective_field_goal_percentage: dict[str, float],
    penalty_kicks_made: dict[str, int],
    penalty_kicks_attempted: dict[str, int],
    shots_total: dict[str, int],
    shots_on_target: dict[str, int],
    yellow_cards: dict[str, int],
    red_cards: dict[str, int],
    touches: dict[str, int],
    expected_goals: dict[str, float],
    non_penalty_expected_goals: dict[str, float],
    expected_assisted_goals: dict[str, float],
    shot_creating_actions: dict[str, int],
    goal_creating_actions: dict[str, int],
    passes_completed: dict[str, int],
    passes_attempted: dict[str, int],
    pass_completion: dict[str, int],
    progressive_passes: dict[str, int],
    carries: dict[str, int],
    progressive_carries: dict[str, int],
    take_ons_attempted: dict[str, int],
    successful_take_ons: dict[str, int],
    total_passing_distance: dict[str, int],
    progressive_passing_distance: dict[str, int],
    passes_completed_short: dict[str, int],
    passes_attempted_short: dict[str, int],
    pass_completion_short: dict[str, int],
    passes_completed_medium: dict[str, int],
    passes_attempted_medium: dict[str, int],
    pass_completion_medium: dict[str, int],
    passes_completed_long: dict[str, int],
    passes_attempted_long: dict[str, int],
    pass_completion_long: dict[str, int],
    expected_assists: dict[str, float],
    key_passes: dict[str, int],
    passes_into_final_third: dict[str, int],
    passes_into_penalty_area: dict[str, int],
    crosses_into_penalty_area: dict[str, int],
    live_ball_passes: dict[str, int],
    dead_ball_passes: dict[str, int],
    passes_from_free_kicks: dict[str, int],
    through_balls: dict[str, int],
    switches: dict[str, int],
    crosses: dict[str, int],
    throw_ins_taken: dict[str, int],
    corner_kicks: dict[str, int],
    inswinging_corner_kicks: dict[str, int],
    outswinging_corner_kicks: dict[str, int],
    straight_corner_kicks: dict[str, int],
    passes_offside: dict[str, int],
    passes_blocked: dict[str, int],
    tackles_won: dict[str, int],
    tackles_in_defensive_third: dict[str, int],
    tackles_in_middle_third: dict[str, int],
    tackles_in_attacking_third: dict[str, int],
    dribblers_tackled: dict[str, int],
    dribbles_challenged: dict[str, int],
    percent_of_dribblers_tackled: dict[str, float],
    challenges_lost: dict[str, int],
    shots_blocked: dict[str, int],
    tackles_plus_interceptions: dict[str, int],
    errors: dict[str, int],
    touches_in_defensive_penalty_area: dict[str, int],
    touches_in_defensive_third: dict[str, int],
    touches_in_middle_third: dict[str, int],
    touches_in_attacking_third: dict[str, int],
    touches_in_attacking_penalty_area: dict[str, int],
    live_ball_touches: dict[str, int],
    successful_take_on_percentage: dict[str, float],
    times_tackled_during_take_ons: dict[str, int],
    tackled_during_take_ons_percentage: dict[str, int],
    total_carrying_distance: dict[str, int],
    progressive_carrying_distance: dict[str, int],
    carries_into_final_third: dict[str, int],
    carries_into_penalty_area: dict[str, int],
    miscontrols: dict[str, int],
    dispossessed: dict[str, int],
    passes_received: dict[str, int],
    progressive_passes_received: dict[str, int],
    second_yellow_card: dict[str, int],
    fouls_committed: dict[str, int],
    fouls_drawn: dict[str, int],
    offsides: dict[str, int],
    penalty_kicks_won: dict[str, int],
    penalty_kicks_conceded: dict[str, int],
    own_goals: dict[str, int],
    ball_recoveries: dict[str, int],
    aerials_won: dict[str, int],
    aerials_lost: dict[str, int],
    percentage_of_aerials_won: dict[str, float],
    shots_on_target_against: dict[str, int],
    post_shot_expected_goals: dict[str, int],
    passes_attempted_minus_goal_kicks: dict[str, int],
    throws_attempted: dict[str, int],
    percentage_of_passes_that_were_launched: dict[str, float],
    average_pass_length: dict[str, float],
    goal_kicks_attempted: dict[str, int],
    percentage_of_goal_kicks_that_were_launched: dict[str, float],
    average_goal_kick_length: dict[str, float],
    crosses_faced: dict[str, int],
    crosses_stopped: dict[str, int],
    percentage_crosses_stopped: dict[str, float],
    defensive_actions_outside_penalty_area: dict[str, int],
    average_distance_of_defensive_actions: dict[str, float],
    three_point_attempt_rate: dict[str, float],
    tackles: dict[str, int],
    interceptions: dict[str, int],
    clearances: dict[str, int],
    free_throw_attempt_rate: dict[str, float],
    offensive_rebound_percentage: dict[str, float],
    defensive_rebound_percentage: dict[str, float],
    total_rebound_percentage: dict[str, float],
    assist_percentage: dict[str, float],
    steal_percentage: dict[str, float],
    block_percentage: dict[str, float],
    turnover_percentage: dict[str, float],
    usage_percentage: dict[str, float],
    offensive_rating: dict[str, int],
    defensive_rating: dict[str, int],
    box_plus_minus: dict[str, float],
) -> PlayerModel | None:
    return _create_sportsreference_player_model(
        session=session,
        player_url=player_url,
        fg=fg,
        fga=fga,
        offensive_rebounds=offensive_rebounds,
        assists=assists,
        turnovers=turnovers,
        positions=positions,
        positions_validator=positions_validator,
        sex=sex,
        dt=dt,
        minutes_played=minutes_played,
        three_point_field_goals=three_point_field_goals,
        three_point_field_goals_attempted=three_point_field_goals_attempted,
        free_throws=free_throws,
        free_throws_attempted=free_throws_attempted,
        defensive_rebounds=defensive_rebounds,
        steals=steals,
        blocks=blocks,
        personal_fouls=personal_fouls,
        points=points,
        game_scores=game_scores,
        point_differentials=point_differentials,
        version=version,
        goals=goals,
        penalties_in_minutes=penalties_in_minutes,
        even_strength_goals=even_strength_goals,
        power_play_goals=power_play_goals,
        short_handed_goals=short_handed_goals,
        game_winning_goals=game_winning_goals,
        even_strength_assists=even_strength_assists,
        power_play_assists=power_play_assists,
        short_handed_assists=short_handed_assists,
        shots_on_goal=shots_on_goal,
        shooting_percentage=shooting_percentage,
        shifts=shifts,
        time_on_ice=time_on_ice,
        decision=decision,
        goals_against=goals_against,
        shots_against=shots_against,
        saves=saves,
        save_percentage=save_percentage,
        shutouts=shutouts,
        individual_corsi_for_events=individual_corsi_for_events,
        on_shot_ice_for_events=on_shot_ice_for_events,
        on_shot_ice_against_events=on_shot_ice_against_events,
        corsi_for_percentage=corsi_for_percentage,
        relative_corsi_for_percentage=relative_corsi_for_percentage,
        offensive_zone_starts=offensive_zone_starts,
        defensive_zone_starts=defensive_zone_starts,
        offensive_zone_start_percentage=offensive_zone_start_percentage,
        hits=hits,
        true_shooting_percentage=true_shooting_percentage,
        at_bats=at_bats,
        runs_scored=runs_scored,
        runs_batted_in=runs_batted_in,
        bases_on_balls=bases_on_balls,
        strikeouts=strikeouts,
        plate_appearances=plate_appearances,
        hits_at_bats=hits_at_bats,
        obp=obp,
        slg=slg,
        ops=ops,
        pitches=pitches,
        strikes=strikes,
        win_probability_added=win_probability_added,
        average_leverage_index=average_leverage_index,
        wpa_plus=wpa_plus,
        wpa_minus=wpa_minus,
        cwpa=cwpa,
        acli=acli,
        re24=re24,
        putouts=putouts,
        innings_pitched=innings_pitched,
        earned_runs=earned_runs,
        home_runs=home_runs,
        era=era,
        batters_faced=batters_faced,
        strikes_by_contact=strikes_by_contact,
        strikes_swinging=strikes_swinging,
        strikes_looking=strikes_looking,
        ground_balls=ground_balls,
        fly_balls=fly_balls,
        line_drives=line_drives,
        inherited_runners=inherited_runners,
        inherited_scores=inherited_scores,
        effective_field_goal_percentage=effective_field_goal_percentage,
        penalty_kicks_made=penalty_kicks_made,
        penalty_kicks_attempted=penalty_kicks_attempted,
        shots_total=shots_total,
        shots_on_target=shots_on_target,
        yellow_cards=yellow_cards,
        red_cards=red_cards,
        touches=touches,
        expected_goals=expected_goals,
        non_penalty_expected_goals=non_penalty_expected_goals,
        expected_assisted_goals=expected_assisted_goals,
        shot_creating_actions=shot_creating_actions,
        goal_creating_actions=goal_creating_actions,
        passes_completed=passes_completed,
        passes_attempted=passes_attempted,
        pass_completion=pass_completion,
        progressive_passes=progressive_passes,
        carries=carries,
        progressive_carries=progressive_carries,
        take_ons_attempted=take_ons_attempted,
        successful_take_ons=successful_take_ons,
        total_passing_distance=total_passing_distance,
        progressive_passing_distance=progressive_passing_distance,
        passes_completed_short=passes_completed_short,
        passes_attempted_short=passes_attempted_short,
        pass_completion_short=pass_completion_short,
        passes_completed_medium=passes_completed_medium,
        passes_attempted_medium=passes_attempted_medium,
        pass_completion_medium=pass_completion_medium,
        passes_completed_long=passes_completed_long,
        passes_attempted_long=passes_attempted_long,
        pass_completion_long=pass_completion_long,
        expected_assists=expected_assists,
        key_passes=key_passes,
        passes_into_final_third=passes_into_final_third,
        passes_into_penalty_area=passes_into_penalty_area,
        crosses_into_penalty_area=crosses_into_penalty_area,
        live_ball_passes=live_ball_passes,
        dead_ball_passes=dead_ball_passes,
        passes_from_free_kicks=passes_from_free_kicks,
        through_balls=through_balls,
        switches=switches,
        crosses=crosses,
        throw_ins_taken=throw_ins_taken,
        corner_kicks=corner_kicks,
        inswinging_corner_kicks=inswinging_corner_kicks,
        outswinging_corner_kicks=outswinging_corner_kicks,
        straight_corner_kicks=straight_corner_kicks,
        passes_offside=passes_offside,
        passes_blocked=passes_blocked,
        tackles_won=tackles_won,
        tackles_in_defensive_third=tackles_in_defensive_third,
        tackles_in_middle_third=tackles_in_middle_third,
        tackles_in_attacking_third=tackles_in_attacking_third,
        dribblers_tackled=dribblers_tackled,
        dribbles_challenged=dribbles_challenged,
        percent_of_dribblers_tackled=percent_of_dribblers_tackled,
        challenges_lost=challenges_lost,
        shots_blocked=shots_blocked,
        tackles_plus_interceptions=tackles_plus_interceptions,
        errors=errors,
        touches_in_defensive_penalty_area=touches_in_defensive_penalty_area,
        touches_in_defensive_third=touches_in_defensive_third,
        touches_in_middle_third=touches_in_middle_third,
        touches_in_attacking_third=touches_in_attacking_third,
        touches_in_attacking_penalty_area=touches_in_attacking_penalty_area,
        live_ball_touches=live_ball_touches,
        successful_take_on_percentage=successful_take_on_percentage,
        times_tackled_during_take_ons=times_tackled_during_take_ons,
        tackled_during_take_ons_percentage=tackled_during_take_ons_percentage,
        total_carrying_distance=total_carrying_distance,
        progressive_carrying_distance=progressive_carrying_distance,
        carries_into_final_third=carries_into_final_third,
        carries_into_penalty_area=carries_into_penalty_area,
        miscontrols=miscontrols,
        dispossessed=dispossessed,
        passes_received=passes_received,
        progressive_passes_received=progressive_passes_received,
        second_yellow_card=second_yellow_card,
        fouls_committed=fouls_committed,
        fouls_drawn=fouls_drawn,
        offsides=offsides,
        penalty_kicks_won=penalty_kicks_won,
        penalty_kicks_conceded=penalty_kicks_conceded,
        own_goals=own_goals,
        ball_recoveries=ball_recoveries,
        aerials_won=aerials_won,
        aerials_lost=aerials_lost,
        percentage_of_aerials_won=percentage_of_aerials_won,
        shots_on_target_against=shots_on_target_against,
        post_shot_expected_goals=post_shot_expected_goals,
        passes_attempted_minus_goal_kicks=passes_attempted_minus_goal_kicks,
        throws_attempted=throws_attempted,
        percentage_of_passes_that_were_launched=percentage_of_passes_that_were_launched,
        average_pass_length=average_pass_length,
        goal_kicks_attempted=goal_kicks_attempted,
        percentage_of_goal_kicks_that_were_launched=percentage_of_goal_kicks_that_were_launched,
        average_goal_kick_length=average_goal_kick_length,
        crosses_faced=crosses_faced,
        crosses_stopped=crosses_stopped,
        percentage_crosses_stopped=percentage_crosses_stopped,
        defensive_actions_outside_penalty_area=defensive_actions_outside_penalty_area,
        average_distance_of_defensive_actions=average_distance_of_defensive_actions,
        three_point_attempt_rate=three_point_attempt_rate,
        tackles=tackles,
        interceptions=interceptions,
        clearances=clearances,
        free_throw_attempt_rate=free_throw_attempt_rate,
        offensive_rebound_percentage=offensive_rebound_percentage,
        defensive_rebound_percentage=defensive_rebound_percentage,
        total_rebound_percentage=total_rebound_percentage,
        assist_percentage=assist_percentage,
        steal_percentage=steal_percentage,
        block_percentage=block_percentage,
        turnover_percentage=turnover_percentage,
        usage_percentage=usage_percentage,
        offensive_rating=offensive_rating,
        defensive_rating=defensive_rating,
        box_plus_minus=box_plus_minus,
    )


def create_sportsreference_player_model(
    session: ScrapeSession,
    player_url: str,
    fg: dict[str, int],
    fga: dict[str, int],
    offensive_rebounds: dict[str, int],
    assists: dict[str, int],
    turnovers: dict[str, int],
    positions: dict[str, str],
    positions_validator: dict[str, str],
    sex: Sex,
    dt: datetime.datetime,
    minutes_played: dict[str, datetime.timedelta],
    three_point_field_goals: dict[str, int],
    three_point_field_goals_attempted: dict[str, int],
    free_throws: dict[str, int],
    free_throws_attempted: dict[str, int],
    defensive_rebounds: dict[str, int],
    steals: dict[str, int],
    blocks: dict[str, int],
    personal_fouls: dict[str, int],
    points: dict[str, int],
    game_scores: dict[str, float],
    point_differentials: dict[str, int],
    goals: dict[str, int],
    penalties_in_minutes: dict[str, datetime.timedelta],
    even_strength_goals: dict[str, int],
    power_play_goals: dict[str, int],
    short_handed_goals: dict[str, int],
    game_winning_goals: dict[str, int],
    even_strength_assists: dict[str, int],
    power_play_assists: dict[str, int],
    short_handed_assists: dict[str, int],
    shots_on_goal: dict[str, int],
    shooting_percentage: dict[str, float],
    shifts: dict[str, int],
    time_on_ice: dict[str, datetime.timedelta],
    decision: dict[str, str],
    goals_against: dict[str, int],
    shots_against: dict[str, int],
    saves: dict[str, int],
    save_percentage: dict[str, float],
    shutouts: dict[str, int],
    individual_corsi_for_events: dict[str, int],
    on_shot_ice_for_events: dict[str, int],
    on_shot_ice_against_events: dict[str, int],
    corsi_for_percentage: dict[str, float],
    relative_corsi_for_percentage: dict[str, float],
    offensive_zone_starts: dict[str, int],
    defensive_zone_starts: dict[str, int],
    offensive_zone_start_percentage: dict[str, float],
    hits: dict[str, int],
    true_shooting_percentage: dict[str, float],
    at_bats: dict[str, int],
    runs_scored: dict[str, int],
    runs_batted_in: dict[str, int],
    bases_on_balls: dict[str, int],
    strikeouts: dict[str, int],
    plate_appearances: dict[str, int],
    hits_at_bats: dict[str, float],
    obp: dict[str, float],
    slg: dict[str, float],
    ops: dict[str, float],
    pitches: dict[str, int],
    strikes: dict[str, int],
    win_probability_added: dict[str, float],
    average_leverage_index: dict[str, float],
    wpa_plus: dict[str, float],
    wpa_minus: dict[str, float],
    cwpa: dict[str, float],
    acli: dict[str, float],
    re24: dict[str, float],
    putouts: dict[str, int],
    innings_pitched: dict[str, int],
    earned_runs: dict[str, int],
    home_runs: dict[str, int],
    era: dict[str, float],
    batters_faced: dict[str, int],
    strikes_by_contact: dict[str, int],
    strikes_swinging: dict[str, int],
    strikes_looking: dict[str, int],
    ground_balls: dict[str, int],
    fly_balls: dict[str, int],
    line_drives: dict[str, int],
    inherited_runners: dict[str, int],
    inherited_scores: dict[str, int],
    effective_field_goal_percentage: dict[str, float],
    penalty_kicks_made: dict[str, int],
    penalty_kicks_attempted: dict[str, int],
    shots_total: dict[str, int],
    shots_on_target: dict[str, int],
    yellow_cards: dict[str, int],
    red_cards: dict[str, int],
    touches: dict[str, int],
    expected_goals: dict[str, float],
    non_penalty_expected_goals: dict[str, float],
    expected_assisted_goals: dict[str, float],
    shot_creating_actions: dict[str, int],
    goal_creating_actions: dict[str, int],
    passes_completed: dict[str, int],
    passes_attempted: dict[str, int],
    pass_completion: dict[str, int],
    progressive_passes: dict[str, int],
    carries: dict[str, int],
    progressive_carries: dict[str, int],
    take_ons_attempted: dict[str, int],
    successful_take_ons: dict[str, int],
    total_passing_distance: dict[str, int],
    progressive_passing_distance: dict[str, int],
    passes_completed_short: dict[str, int],
    passes_attempted_short: dict[str, int],
    pass_completion_short: dict[str, int],
    passes_completed_medium: dict[str, int],
    passes_attempted_medium: dict[str, int],
    pass_completion_medium: dict[str, int],
    passes_completed_long: dict[str, int],
    passes_attempted_long: dict[str, int],
    pass_completion_long: dict[str, int],
    expected_assists: dict[str, float],
    key_passes: dict[str, int],
    passes_into_final_third: dict[str, int],
    passes_into_penalty_area: dict[str, int],
    crosses_into_penalty_area: dict[str, int],
    live_ball_passes: dict[str, int],
    dead_ball_passes: dict[str, int],
    passes_from_free_kicks: dict[str, int],
    through_balls: dict[str, int],
    switches: dict[str, int],
    crosses: dict[str, int],
    throw_ins_taken: dict[str, int],
    corner_kicks: dict[str, int],
    inswinging_corner_kicks: dict[str, int],
    outswinging_corner_kicks: dict[str, int],
    straight_corner_kicks: dict[str, int],
    passes_offside: dict[str, int],
    passes_blocked: dict[str, int],
    tackles_won: dict[str, int],
    tackles_in_defensive_third: dict[str, int],
    tackles_in_middle_third: dict[str, int],
    tackles_in_attacking_third: dict[str, int],
    dribblers_tackled: dict[str, int],
    dribbles_challenged: dict[str, int],
    percent_of_dribblers_tackled: dict[str, float],
    challenges_lost: dict[str, int],
    shots_blocked: dict[str, int],
    tackles_plus_interceptions: dict[str, int],
    errors: dict[str, int],
    touches_in_defensive_penalty_area: dict[str, int],
    touches_in_defensive_third: dict[str, int],
    touches_in_middle_third: dict[str, int],
    touches_in_attacking_third: dict[str, int],
    touches_in_attacking_penalty_area: dict[str, int],
    live_ball_touches: dict[str, int],
    successful_take_on_percentage: dict[str, float],
    times_tackled_during_take_ons: dict[str, int],
    tackled_during_take_ons_percentage: dict[str, int],
    total_carrying_distance: dict[str, int],
    progressive_carrying_distance: dict[str, int],
    carries_into_final_third: dict[str, int],
    carries_into_penalty_area: dict[str, int],
    miscontrols: dict[str, int],
    dispossessed: dict[str, int],
    passes_received: dict[str, int],
    progressive_passes_received: dict[str, int],
    second_yellow_card: dict[str, int],
    fouls_committed: dict[str, int],
    fouls_drawn: dict[str, int],
    offsides: dict[str, int],
    penalty_kicks_won: dict[str, int],
    penalty_kicks_conceded: dict[str, int],
    own_goals: dict[str, int],
    ball_recoveries: dict[str, int],
    aerials_won: dict[str, int],
    aerials_lost: dict[str, int],
    percentage_of_aerials_won: dict[str, float],
    shots_on_target_against: dict[str, int],
    post_shot_expected_goals: dict[str, int],
    passes_attempted_minus_goal_kicks: dict[str, int],
    throws_attempted: dict[str, int],
    percentage_of_passes_that_were_launched: dict[str, float],
    average_pass_length: dict[str, float],
    goal_kicks_attempted: dict[str, int],
    percentage_of_goal_kicks_that_were_launched: dict[str, float],
    average_goal_kick_length: dict[str, float],
    crosses_faced: dict[str, int],
    crosses_stopped: dict[str, int],
    percentage_crosses_stopped: dict[str, float],
    defensive_actions_outside_penalty_area: dict[str, int],
    average_distance_of_defensive_actions: dict[str, float],
    three_point_attempt_rate: dict[str, float],
    tackles: dict[str, int],
    interceptions: dict[str, int],
    clearances: dict[str, int],
    free_throw_attempt_rate: dict[str, float],
    offensive_rebound_percentage: dict[str, float],
    defensive_rebound_percentage: dict[str, float],
    total_rebound_percentage: dict[str, float],
    assist_percentage: dict[str, float],
    steal_percentage: dict[str, float],
    block_percentage: dict[str, float],
    turnover_percentage: dict[str, float],
    usage_percentage: dict[str, float],
    offensive_rating: dict[str, int],
    defensive_rating: dict[str, int],
    box_plus_minus: dict[str, float],
) -> PlayerModel | None:
    """Create a player model from sports reference."""
    if not pytest_is_running.is_running():
        return _cached_create_sportsreference_player_model(
            session=session,
            player_url=player_url,
            fg=fg,
            fga=fga,
            offensive_rebounds=offensive_rebounds,
            assists=assists,
            turnovers=turnovers,
            positions=positions,
            positions_validator=positions_validator,
            sex=sex,
            dt=dt,
            minutes_played=minutes_played,
            three_point_field_goals=three_point_field_goals,
            three_point_field_goals_attempted=three_point_field_goals_attempted,
            free_throws=free_throws,
            free_throws_attempted=free_throws_attempted,
            defensive_rebounds=defensive_rebounds,
            steals=steals,
            blocks=blocks,
            personal_fouls=personal_fouls,
            points=points,
            game_scores=game_scores,
            point_differentials=point_differentials,
            version=VERSION,
            goals=goals,
            penalties_in_minutes=penalties_in_minutes,
            even_strength_goals=even_strength_goals,
            power_play_goals=power_play_goals,
            short_handed_goals=short_handed_goals,
            game_winning_goals=game_winning_goals,
            even_strength_assists=even_strength_assists,
            power_play_assists=power_play_assists,
            short_handed_assists=short_handed_assists,
            shots_on_goal=shots_on_goal,
            shooting_percentage=shooting_percentage,
            shifts=shifts,
            time_on_ice=time_on_ice,
            decision=decision,
            goals_against=goals_against,
            shots_against=shots_against,
            saves=saves,
            save_percentage=save_percentage,
            shutouts=shutouts,
            individual_corsi_for_events=individual_corsi_for_events,
            on_shot_ice_for_events=on_shot_ice_for_events,
            on_shot_ice_against_events=on_shot_ice_against_events,
            corsi_for_percentage=corsi_for_percentage,
            relative_corsi_for_percentage=relative_corsi_for_percentage,
            offensive_zone_starts=offensive_zone_starts,
            defensive_zone_starts=defensive_zone_starts,
            offensive_zone_start_percentage=offensive_zone_start_percentage,
            hits=hits,
            true_shooting_percentage=true_shooting_percentage,
            at_bats=at_bats,
            runs_scored=runs_scored,
            runs_batted_in=runs_batted_in,
            bases_on_balls=bases_on_balls,
            strikeouts=strikeouts,
            plate_appearances=plate_appearances,
            hits_at_bats=hits_at_bats,
            obp=obp,
            slg=slg,
            ops=ops,
            pitches=pitches,
            strikes=strikes,
            win_probability_added=win_probability_added,
            average_leverage_index=average_leverage_index,
            wpa_plus=wpa_plus,
            wpa_minus=wpa_minus,
            cwpa=cwpa,
            acli=acli,
            re24=re24,
            putouts=putouts,
            innings_pitched=innings_pitched,
            earned_runs=earned_runs,
            home_runs=home_runs,
            era=era,
            batters_faced=batters_faced,
            strikes_by_contact=strikes_by_contact,
            strikes_swinging=strikes_swinging,
            strikes_looking=strikes_looking,
            ground_balls=ground_balls,
            fly_balls=fly_balls,
            line_drives=line_drives,
            inherited_runners=inherited_runners,
            inherited_scores=inherited_scores,
            effective_field_goal_percentage=effective_field_goal_percentage,
            penalty_kicks_made=penalty_kicks_made,
            penalty_kicks_attempted=penalty_kicks_attempted,
            shots_total=shots_total,
            shots_on_target=shots_on_target,
            yellow_cards=yellow_cards,
            red_cards=red_cards,
            touches=touches,
            expected_goals=expected_goals,
            non_penalty_expected_goals=non_penalty_expected_goals,
            expected_assisted_goals=expected_assisted_goals,
            shot_creating_actions=shot_creating_actions,
            goal_creating_actions=goal_creating_actions,
            passes_completed=passes_completed,
            passes_attempted=passes_attempted,
            pass_completion=pass_completion,
            progressive_passes=progressive_passes,
            carries=carries,
            progressive_carries=progressive_carries,
            take_ons_attempted=take_ons_attempted,
            successful_take_ons=successful_take_ons,
            total_passing_distance=total_passing_distance,
            progressive_passing_distance=progressive_passing_distance,
            passes_completed_short=passes_completed_short,
            passes_attempted_short=passes_attempted_short,
            pass_completion_short=pass_completion_short,
            passes_completed_medium=passes_completed_medium,
            passes_attempted_medium=passes_attempted_medium,
            pass_completion_medium=pass_completion_medium,
            passes_completed_long=passes_completed_long,
            passes_attempted_long=passes_attempted_long,
            pass_completion_long=pass_completion_long,
            expected_assists=expected_assists,
            key_passes=key_passes,
            passes_into_final_third=passes_into_final_third,
            passes_into_penalty_area=passes_into_penalty_area,
            crosses_into_penalty_area=crosses_into_penalty_area,
            live_ball_passes=live_ball_passes,
            dead_ball_passes=dead_ball_passes,
            passes_from_free_kicks=passes_from_free_kicks,
            through_balls=through_balls,
            switches=switches,
            crosses=crosses,
            throw_ins_taken=throw_ins_taken,
            corner_kicks=corner_kicks,
            inswinging_corner_kicks=inswinging_corner_kicks,
            outswinging_corner_kicks=outswinging_corner_kicks,
            straight_corner_kicks=straight_corner_kicks,
            passes_offside=passes_offside,
            passes_blocked=passes_blocked,
            tackles_won=tackles_won,
            tackles_in_defensive_third=tackles_in_defensive_third,
            tackles_in_middle_third=tackles_in_middle_third,
            tackles_in_attacking_third=tackles_in_attacking_third,
            dribblers_tackled=dribblers_tackled,
            dribbles_challenged=dribbles_challenged,
            percent_of_dribblers_tackled=percent_of_dribblers_tackled,
            challenges_lost=challenges_lost,
            shots_blocked=shots_blocked,
            tackles_plus_interceptions=tackles_plus_interceptions,
            errors=errors,
            touches_in_defensive_penalty_area=touches_in_defensive_penalty_area,
            touches_in_defensive_third=touches_in_defensive_third,
            touches_in_middle_third=touches_in_middle_third,
            touches_in_attacking_third=touches_in_attacking_third,
            touches_in_attacking_penalty_area=touches_in_attacking_penalty_area,
            live_ball_touches=live_ball_touches,
            successful_take_on_percentage=successful_take_on_percentage,
            times_tackled_during_take_ons=times_tackled_during_take_ons,
            tackled_during_take_ons_percentage=tackled_during_take_ons_percentage,
            total_carrying_distance=total_carrying_distance,
            progressive_carrying_distance=progressive_carrying_distance,
            carries_into_final_third=carries_into_final_third,
            carries_into_penalty_area=carries_into_penalty_area,
            miscontrols=miscontrols,
            dispossessed=dispossessed,
            passes_received=passes_received,
            progressive_passes_received=progressive_passes_received,
            second_yellow_card=second_yellow_card,
            fouls_committed=fouls_committed,
            fouls_drawn=fouls_drawn,
            offsides=offsides,
            penalty_kicks_won=penalty_kicks_won,
            penalty_kicks_conceded=penalty_kicks_conceded,
            own_goals=own_goals,
            ball_recoveries=ball_recoveries,
            aerials_won=aerials_won,
            aerials_lost=aerials_lost,
            percentage_of_aerials_won=percentage_of_aerials_won,
            shots_on_target_against=shots_on_target_against,
            post_shot_expected_goals=post_shot_expected_goals,
            passes_attempted_minus_goal_kicks=passes_attempted_minus_goal_kicks,
            throws_attempted=throws_attempted,
            percentage_of_passes_that_were_launched=percentage_of_passes_that_were_launched,
            average_pass_length=average_pass_length,
            goal_kicks_attempted=goal_kicks_attempted,
            percentage_of_goal_kicks_that_were_launched=percentage_of_goal_kicks_that_were_launched,
            average_goal_kick_length=average_goal_kick_length,
            crosses_faced=crosses_faced,
            crosses_stopped=crosses_stopped,
            percentage_crosses_stopped=percentage_crosses_stopped,
            defensive_actions_outside_penalty_area=defensive_actions_outside_penalty_area,
            average_distance_of_defensive_actions=average_distance_of_defensive_actions,
            three_point_attempt_rate=three_point_attempt_rate,
            tackles=tackles,
            interceptions=interceptions,
            clearances=clearances,
            free_throw_attempt_rate=free_throw_attempt_rate,
            offensive_rebound_percentage=offensive_rebound_percentage,
            defensive_rebound_percentage=defensive_rebound_percentage,
            total_rebound_percentage=total_rebound_percentage,
            assist_percentage=assist_percentage,
            steal_percentage=steal_percentage,
            block_percentage=block_percentage,
            turnover_percentage=turnover_percentage,
            usage_percentage=usage_percentage,
            offensive_rating=offensive_rating,
            defensive_rating=defensive_rating,
            box_plus_minus=box_plus_minus,
        )
    with session.cache_disabled():
        return _create_sportsreference_player_model(
            session=session,
            player_url=player_url,
            fg=fg,
            fga=fga,
            offensive_rebounds=offensive_rebounds,
            assists=assists,
            turnovers=turnovers,
            positions=positions,
            positions_validator=positions_validator,
            sex=sex,
            dt=dt,
            minutes_played=minutes_played,
            three_point_field_goals=three_point_field_goals,
            three_point_field_goals_attempted=three_point_field_goals_attempted,
            free_throws=free_throws,
            free_throws_attempted=free_throws_attempted,
            defensive_rebounds=defensive_rebounds,
            steals=steals,
            blocks=blocks,
            personal_fouls=personal_fouls,
            points=points,
            game_scores=game_scores,
            point_differentials=point_differentials,
            version=VERSION,
            goals=goals,
            penalties_in_minutes=penalties_in_minutes,
            even_strength_goals=even_strength_goals,
            power_play_goals=power_play_goals,
            short_handed_goals=short_handed_goals,
            game_winning_goals=game_winning_goals,
            even_strength_assists=even_strength_assists,
            power_play_assists=power_play_assists,
            short_handed_assists=short_handed_assists,
            shots_on_goal=shots_on_goal,
            shooting_percentage=shooting_percentage,
            shifts=shifts,
            time_on_ice=time_on_ice,
            decision=decision,
            goals_against=goals_against,
            shots_against=shots_against,
            saves=saves,
            save_percentage=save_percentage,
            shutouts=shutouts,
            individual_corsi_for_events=individual_corsi_for_events,
            on_shot_ice_for_events=on_shot_ice_for_events,
            on_shot_ice_against_events=on_shot_ice_against_events,
            corsi_for_percentage=corsi_for_percentage,
            relative_corsi_for_percentage=relative_corsi_for_percentage,
            offensive_zone_starts=offensive_zone_starts,
            defensive_zone_starts=defensive_zone_starts,
            offensive_zone_start_percentage=offensive_zone_start_percentage,
            hits=hits,
            true_shooting_percentage=true_shooting_percentage,
            at_bats=at_bats,
            runs_scored=runs_scored,
            runs_batted_in=runs_batted_in,
            bases_on_balls=bases_on_balls,
            strikeouts=strikeouts,
            plate_appearances=plate_appearances,
            hits_at_bats=hits_at_bats,
            obp=obp,
            slg=slg,
            ops=ops,
            pitches=pitches,
            strikes=strikes,
            win_probability_added=win_probability_added,
            average_leverage_index=average_leverage_index,
            wpa_plus=wpa_plus,
            wpa_minus=wpa_minus,
            cwpa=cwpa,
            acli=acli,
            re24=re24,
            putouts=putouts,
            innings_pitched=innings_pitched,
            earned_runs=earned_runs,
            home_runs=home_runs,
            era=era,
            batters_faced=batters_faced,
            strikes_by_contact=strikes_by_contact,
            strikes_swinging=strikes_swinging,
            strikes_looking=strikes_looking,
            ground_balls=ground_balls,
            fly_balls=fly_balls,
            line_drives=line_drives,
            inherited_runners=inherited_runners,
            inherited_scores=inherited_scores,
            effective_field_goal_percentage=effective_field_goal_percentage,
            penalty_kicks_made=penalty_kicks_made,
            penalty_kicks_attempted=penalty_kicks_attempted,
            shots_total=shots_total,
            shots_on_target=shots_on_target,
            yellow_cards=yellow_cards,
            red_cards=red_cards,
            touches=touches,
            expected_goals=expected_goals,
            non_penalty_expected_goals=non_penalty_expected_goals,
            expected_assisted_goals=expected_assisted_goals,
            shot_creating_actions=shot_creating_actions,
            goal_creating_actions=goal_creating_actions,
            passes_completed=passes_completed,
            passes_attempted=passes_attempted,
            pass_completion=pass_completion,
            progressive_passes=progressive_passes,
            carries=carries,
            progressive_carries=progressive_carries,
            take_ons_attempted=take_ons_attempted,
            successful_take_ons=successful_take_ons,
            total_passing_distance=total_passing_distance,
            progressive_passing_distance=progressive_passing_distance,
            passes_completed_short=passes_completed_short,
            passes_attempted_short=passes_attempted_short,
            pass_completion_short=pass_completion_short,
            passes_completed_medium=passes_completed_medium,
            passes_attempted_medium=passes_attempted_medium,
            pass_completion_medium=pass_completion_medium,
            passes_completed_long=passes_completed_long,
            passes_attempted_long=passes_attempted_long,
            pass_completion_long=pass_completion_long,
            expected_assists=expected_assists,
            key_passes=key_passes,
            passes_into_final_third=passes_into_final_third,
            passes_into_penalty_area=passes_into_penalty_area,
            crosses_into_penalty_area=crosses_into_penalty_area,
            live_ball_passes=live_ball_passes,
            dead_ball_passes=dead_ball_passes,
            passes_from_free_kicks=passes_from_free_kicks,
            through_balls=through_balls,
            switches=switches,
            crosses=crosses,
            throw_ins_taken=throw_ins_taken,
            corner_kicks=corner_kicks,
            inswinging_corner_kicks=inswinging_corner_kicks,
            outswinging_corner_kicks=outswinging_corner_kicks,
            straight_corner_kicks=straight_corner_kicks,
            passes_offside=passes_offside,
            passes_blocked=passes_blocked,
            tackles_won=tackles_won,
            tackles_in_defensive_third=tackles_in_defensive_third,
            tackles_in_middle_third=tackles_in_middle_third,
            tackles_in_attacking_third=tackles_in_attacking_third,
            dribblers_tackled=dribblers_tackled,
            dribbles_challenged=dribbles_challenged,
            percent_of_dribblers_tackled=percent_of_dribblers_tackled,
            challenges_lost=challenges_lost,
            shots_blocked=shots_blocked,
            tackles_plus_interceptions=tackles_plus_interceptions,
            errors=errors,
            touches_in_defensive_penalty_area=touches_in_defensive_penalty_area,
            touches_in_defensive_third=touches_in_defensive_third,
            touches_in_middle_third=touches_in_middle_third,
            touches_in_attacking_third=touches_in_attacking_third,
            touches_in_attacking_penalty_area=touches_in_attacking_penalty_area,
            live_ball_touches=live_ball_touches,
            successful_take_on_percentage=successful_take_on_percentage,
            times_tackled_during_take_ons=times_tackled_during_take_ons,
            tackled_during_take_ons_percentage=tackled_during_take_ons_percentage,
            total_carrying_distance=total_carrying_distance,
            progressive_carrying_distance=progressive_carrying_distance,
            carries_into_final_third=carries_into_final_third,
            carries_into_penalty_area=carries_into_penalty_area,
            miscontrols=miscontrols,
            dispossessed=dispossessed,
            passes_received=passes_received,
            progressive_passes_received=progressive_passes_received,
            second_yellow_card=second_yellow_card,
            fouls_committed=fouls_committed,
            fouls_drawn=fouls_drawn,
            offsides=offsides,
            penalty_kicks_won=penalty_kicks_won,
            penalty_kicks_conceded=penalty_kicks_conceded,
            own_goals=own_goals,
            ball_recoveries=ball_recoveries,
            aerials_won=aerials_won,
            aerials_lost=aerials_lost,
            percentage_of_aerials_won=percentage_of_aerials_won,
            shots_on_target_against=shots_on_target_against,
            post_shot_expected_goals=post_shot_expected_goals,
            passes_attempted_minus_goal_kicks=passes_attempted_minus_goal_kicks,
            throws_attempted=throws_attempted,
            percentage_of_passes_that_were_launched=percentage_of_passes_that_were_launched,
            average_pass_length=average_pass_length,
            goal_kicks_attempted=goal_kicks_attempted,
            percentage_of_goal_kicks_that_were_launched=percentage_of_goal_kicks_that_were_launched,
            average_goal_kick_length=average_goal_kick_length,
            crosses_faced=crosses_faced,
            crosses_stopped=crosses_stopped,
            percentage_crosses_stopped=percentage_crosses_stopped,
            defensive_actions_outside_penalty_area=defensive_actions_outside_penalty_area,
            average_distance_of_defensive_actions=average_distance_of_defensive_actions,
            three_point_attempt_rate=three_point_attempt_rate,
            tackles=tackles,
            interceptions=interceptions,
            clearances=clearances,
            free_throw_attempt_rate=free_throw_attempt_rate,
            offensive_rebound_percentage=offensive_rebound_percentage,
            defensive_rebound_percentage=defensive_rebound_percentage,
            total_rebound_percentage=total_rebound_percentage,
            assist_percentage=assist_percentage,
            steal_percentage=steal_percentage,
            block_percentage=block_percentage,
            turnover_percentage=turnover_percentage,
            usage_percentage=usage_percentage,
            offensive_rating=offensive_rating,
            defensive_rating=defensive_rating,
            box_plus_minus=box_plus_minus,
        )



"""ESPNCricInfo game model."""

# pylint: disable=duplicate-code,too-many-locals
import json

import pytest_is_running
from bs4 import BeautifulSoup
from dateutil.parser import parse
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ....cache import MEMORY
from ...game_model import VERSION, GameModel
from ...league import League
from .ipl_espncricinfo_team_model import create_espncricinfo_team_model
from .ipl_espncricinfo_venue_model import create_espncricinfo_venue_model

_LEAGUE_NAMES = {
    "IPL": League.IPL,
}


def _create_espncricinfo_game_model(
    session: ScrapeSession,
    url: str,
    league: League,
    positions_validator: dict[str, str],
    version: str,
) -> GameModel | None:
    response = session.get(url)
    response.raise_for_status()
    soup = BeautifulSoup(response.text, "lxml")
    for script in soup.find_all("script", id="__NEXT_DATA__"):
        data = json.loads(script.get_text())
        data_dict = data["props"]["appPageProps"]["data"]
        game = data_dict["match"]
        if league != _LEAGUE_NAMES.get(game["series"]["name"]):
            return None
        dt = parse(game["startTime"])
        end_dt = parse(game["endTime"])
        grounds = game["grounds"]
        content = data_dict["content"]
        teams_players = content["matchPlayers"]["teamPlayers"]
        innings = content["innings"]
        teams = []
        for team in sorted(game["teams"], key=lambda x: x["isHome"], reverse=True):
            teams.append(
                create_espncricinfo_team_model(
                    session=session,
                    dt=dt,
                    league=league,
                    team=team,
                    positions_validator=positions_validator,
                    teams_players=teams_players,
                    innings=innings,
                    url=url,
                )
            )
        return GameModel(
            dt=dt,
            week=None,
            game_number=None,
            venue=create_espncricinfo_venue_model(grounds, session, dt),
            teams=teams,
            league=str(league),
            year=dt.year,
            season_type=None,
            end_dt=end_dt,
            attendance=None,
            postponed=None,
            play_off=None,
            distance=None,
            dividends=[],
            pot=None,
            version=version,
            umpires=[],
        )
    return None


@MEMORY.cache(ignore=["session"])
def _cached_create_espncricinfo_game_model(
    session: ScrapeSession,
    url: str,
    league: League,
    positions_validator: dict[str, str],
    version: str,
) -> GameModel | None:
    return _create_espncricinfo_game_model(
        session=session,
        url=url,
        league=league,
        positions_validator=positions_validator,
        version=version,
    )


def create_espncricinfo_game_model(
    session: ScrapeSession,
    url: str,
    league: League,
    positions_validator: dict[str, str],
) -> GameModel | None:
    """Create a sports reference game model."""
    if not pytest_is_running.is_running():
        return _cached_create_espncricinfo_game_model(
            session=session,
            url=url,
            league=league,
            positions_validator=positions_validator,
            version=VERSION,
        )
    with session.cache_disabled():
        return _create_espncricinfo_game_model(
            session=session,
            url=url,
            league=league,
            positions_validator=positions_validator,
            version=VERSION,
        )



"""ESPN league model."""

# pylint: disable=too-many-locals,too-many-arguments,line-too-long,too-many-branches,too-many-statements
import datetime
import logging
from typing import Any, Iterator
from urllib.parse import urlparse

import requests
import tqdm
from dateutil.parser import parse
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ..game_model import GameModel
from ..league import League
from ..league_model import SHUTDOWN_FLAG, LeagueModel, needs_shutdown
from ..season_type import SeasonType
from .espn_game_model import create_espn_game_model


def _season_type_from_name(name: str) -> SeasonType:
    if (
        name == "Regular Season"
        or "English Premier League" in name
        or "Group Stage" in name
        or "Spring Regular Season" in name
        or "Round of" in name
        or "Second Round" in name
    ):
        return SeasonType.REGULAR
    if name == "Preseason" or "Spring Training" in name:
        return SeasonType.PRESEASON
    if (
        name == "Postseason"
        or "Quarterfinals" in name
        or "Semifinals" in name
        or "3rd-Place" in name
        or "Final" in name
        or "Third Place" in name
        or "Quarter-finals" in name
        or "Semi-finals" in name
    ):
        return SeasonType.POSTSEASON
    if name == "Off Season":
        return SeasonType.OFFSEASON
    raise ValueError(f"Unrecognised season name: {name}")


class ESPNLeagueModel(LeagueModel):
    """ESPN implementation of the league model."""

    def __init__(
        self,
        start_url: str,
        league: League,
        session: ScrapeSession,
        position: int | None = None,
    ) -> None:
        super().__init__(league, session, position=position)
        self._start_url = start_url

    @classmethod
    def name(cls) -> str:
        return "espn-league-model"

    @classmethod
    def position_validator(cls) -> dict[str, str]:
        """A dictionary that contains the mapping from positions to standard positions."""
        raise NotImplementedError(
            "position_validator is not implemented by parent class"
        )

    def _produce_game(
        self,
        cache_disabled: bool,
        event_item: dict[str, Any],
        week_count: int | None,
        game_number: int,
        season_type_json: dict[str, Any],
        pbar: tqdm.tqdm,
    ) -> Iterator[GameModel]:
        if cache_disabled:
            with self.session.cache_disabled():
                event_response = self.session.get(event_item["$ref"])
        else:
            event_response = self.session.get(event_item["$ref"])
        event_response.raise_for_status()
        event = event_response.json()
        for competition in event["competitions"]:
            game_model = create_espn_game_model(
                competition,
                week_count,
                game_number,
                self.session,
                self.league,
                season_type_json.get("year"),
                _season_type_from_name(season_type_json["name"]),
                self.position_validator(),
            )
            game_number += 1
            pbar.update(1)
            pbar.set_description(
                f"ESPN {game_model.year} - {game_model.season_type} - {game_model.dt}"
            )
            yield game_model

    def _produce_games(
        self,
        week: dict[str, Any],
        week_count: int,
        season_type_json: dict[str, Any],
        pbar: tqdm.tqdm,
        cache_disabled: bool,
    ) -> Iterator[GameModel]:
        events_page = 1
        events_count = 0
        while True:
            if "events" not in week:
                break
            events_response = self.session.get(
                week["events"]["$ref"] + f"&page={events_page}"
            )
            events_response.raise_for_status()
            events = events_response.json()
            for event_item in events["items"]:
                for game_model in self._produce_game(
                    event_item=event_item,
                    game_number=events_count,
                    week_count=week_count,
                    cache_disabled=cache_disabled,
                    season_type_json=season_type_json,
                    pbar=pbar,
                ):
                    events_count += 1
                    yield game_model
            if events_page >= events["pageCount"]:
                break
            events_page += 1
        qbr_page = 1
        qbr_count = 0
        while True:
            if "qbr" not in week:
                break
            qbr_response = self.session.get(week["qbr"]["$ref"] + f"&page={qbr_page}")
            qbr = qbr_response.json()
            for qbr_item in qbr["items"]:
                for game_model in self._produce_game(
                    event_item=qbr_item["event"],
                    game_number=qbr_count,
                    week_count=week_count,
                    cache_disabled=cache_disabled,
                    season_type_json=season_type_json,
                    pbar=pbar,
                ):
                    yield game_model
                    qbr_count += 1
            if qbr_page >= qbr["pageCount"]:
                break
            qbr_page += 1

    def _produce_week_games(
        self,
        season_type_json: dict[str, Any],
        page: int,
        pbar: tqdm.tqdm,
        cache_disabled: bool,
    ) -> Iterator[GameModel]:
        found_pages = False
        if "weeks" in season_type_json:
            game_page = 1
            week_count = 0
            while True:
                if cache_disabled:
                    with self.session.cache_disabled():
                        weeks_response = self.session.get(
                            season_type_json["weeks"]["$ref"] + f"&page={page}"
                        )
                else:
                    weeks_response = self.session.get(
                        season_type_json["weeks"]["$ref"] + f"&page={page}"
                    )
                weeks_response.raise_for_status()
                weeks = weeks_response.json()
                for item in weeks["items"]:
                    if needs_shutdown():
                        return
                    if cache_disabled:
                        with self.session.cache_disabled():
                            week_response = self.session.get(item["$ref"])
                    else:
                        week_response = self.session.get(item["$ref"])
                    week_response.raise_for_status()
                    week = week_response.json()
                    for game_model in self._produce_games(
                        week, week_count, season_type_json, pbar, cache_disabled
                    ):
                        yield game_model
                        found_pages = True
                    week_count += 1
                if game_page >= weeks["pageCount"]:
                    break
                game_page += 1
        if not found_pages and self.league != League.NFL:
            # Lets check the scoreboards via the calendar
            o = urlparse(season_type_json["$ref"])
            path_components = o.path.split("/")
            sport_slug = path_components[3]
            league_slug = path_components[5]
            start_date = parse(season_type_json["startDate"]).date()
            end_date = parse(season_type_json["endDate"]).date()
            calendar_dates = {
                (start_date + datetime.timedelta(days=x))
                for x in range((end_date - start_date).days)
            }
            events_count = 0
            while calendar_dates:
                calendar_date = calendar_dates.pop()
                if calendar_date > datetime.datetime.now().date() - datetime.timedelta(
                    days=2
                ):
                    continue
                dt = calendar_date.strftime("%Y%m%d")
                url = f"https://site.api.espn.com/apis/site/v2/sports/{sport_slug}/{league_slug}/scoreboard?lang=en&region=us&calendartype=whitelist&limit=100&dates={dt}&league={league_slug}"
                scoreboard_response = self.session.get(url)
                scoreboard_response.raise_for_status()
                try:
                    scoreboard = scoreboard_response.json()
                    calendar_list = scoreboard["leagues"][0]["calendar"]
                    if calendar_list:
                        if isinstance(calendar_list[0], str):
                            declared_calendar_dates = {
                                parse(x).date() for x in calendar_list
                            }
                            calendar_dates &= declared_calendar_dates
                    for event in scoreboard["events"]:
                        event_id = event["id"]
                        event_response = self.session.get(
                            f"https://sports.core.api.espn.com/v2/sports/{sport_slug}/leagues/{league_slug}/events/{event_id}?lang=en&region=us"
                        )
                        event_response.raise_for_status()
                        for game_model in self._produce_game(
                            event_item=event_response.json(),
                            game_number=events_count,
                            season_type_json=season_type_json,
                            pbar=pbar,
                            cache_disabled=False,
                            week_count=None,
                        ):
                            yield game_model
                            events_count += 1
                except (requests.exceptions.JSONDecodeError, KeyError) as exc:
                    logging.error(scoreboard_response.text)
                    logging.error(url)
                    logging.error(str(exc))

    @property
    def games(self) -> Iterator[GameModel]:
        try:
            with self.session.wayback_disabled():
                page = 1
                first = True
                with tqdm.tqdm(position=self.position) as pbar:
                    while True:
                        if page == 1:
                            with self.session.cache_disabled():
                                response = self.session.get(
                                    self._start_url + f"&page={page}"
                                )
                        else:
                            response = self.session.get(
                                self._start_url + f"&page={page}"
                            )
                        response.raise_for_status()
                        seasons = response.json()
                        for item in seasons.get("items", []):
                            season_response = self.session.get(item["$ref"])
                            season_response.raise_for_status()
                            season_json = season_response.json()

                            for season_item in season_json["types"]["items"]:
                                season_type_response = self.session.get(
                                    season_item["$ref"]
                                )
                                season_type_response.raise_for_status()
                                season_type_json = season_type_response.json()

                                yield from self._produce_week_games(
                                    season_type_json, page, pbar, first
                                )
                            first = False

                        if page >= seasons.get("pageCount", 0):
                            break
                        page += 1
        except Exception as exc:
            SHUTDOWN_FLAG.set()
            raise exc



"""Tests for the HKJC HKJC game model class."""
import os
import unittest

import requests_mock
from sportsball.data.hkjc.hkjc.hkjc_hkjc_game_model import create_hkjc_hkjc_game_model
from scrapesession.scrapesession import ScrapeSession


class TestHKJCHKJCGameModel(unittest.TestCase):

    def setUp(self):
        self._session = ScrapeSession(backend="memory")
        self._session._wayback_disabled = True
        self.dir = os.path.dirname(__file__)

    def test_invalid_horse_weight(self):
        with requests_mock.Mocker() as m:
            html = ""
            with open(os.path.join(self.dir, "game1.html"), "r") as f:
                html = f.read()
            with open(os.path.join(self.dir, "horse_HK_2024_K204.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Horse/Horse.aspx?HorseID=HK_2024_K204", content=f.read())
            with open(os.path.join(self.dir, "sire_advertise.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/english/Horse/SameSire.aspx?HorseSire=Advertise", content=f.read())
            with open(os.path.join(self.dir, "jockey_CML.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Jockey/JockeyProfile.aspx?JockeyId=CML&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "trainer_MWK.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Trainers/TrainerProfile.aspx?TrainerId=MWK&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "horse_HK_2024_K377.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Horse/Horse.aspx?HorseId=HK_2024_K377", content=f.read())
            with open(os.path.join(self.dir, "sire_land_force.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/english/Horse/SameSire.aspx?HorseSire=Land%20Force", content=f.read())
            with open(os.path.join(self.dir, "jockey_CJE.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Jockey/JockeyProfile.aspx?JockeyId=CJE&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "trainer_YCH.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Trainers/TrainerProfile.aspx?TrainerId=YCH&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "horse_HK_2024_K326.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Horse/Horse.aspx?HorseId=HK_2024_K326", content=f.read())
            with open(os.path.join(self.dir, "sire_pierata.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/english/Horse/SameSire.aspx?HorseSire=Pierata", content=f.read())
            with open(os.path.join(self.dir, "jockey_PZ.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Jockey/JockeyProfile.aspx?JockeyId=PZ&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "trainer_CCW.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Trainers/TrainerProfile.aspx?TrainerId=CCW&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "horse_HK_2024_K165.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Horse/Horse.aspx?HorseId=HK_2024_K165", content=f.read())
            with open(os.path.join(self.dir, "sire_fastnet_rock.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/english/Horse/SameSire.aspx?HorseSire=Fastnet%20Rock", content=f.read())
            with open(os.path.join(self.dir, "jockey_LDE.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Jockey/JockeyProfile.aspx?JockeyId=LDE&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "horse_HK_2024_K406.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Horse/Horse.aspx?HorseId=HK_2024_K406", content=f.read())
            with open(os.path.join(self.dir, "sire_written_tycoon.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/english/Horse/SameSire.aspx?HorseSire=Written%20Tycoon", content=f.read())
            with open(os.path.join(self.dir, "jockey_BHW.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Jockey/JockeyProfile.aspx?JockeyId=BHW&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "trainer_SCS.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Trainers/TrainerProfile.aspx?TrainerId=SCS&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "horse_HK_2024_K433.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Horse/Horse.aspx?HorseId=HK_2024_K433", content=f.read())
            with open(os.path.join(self.dir, "sire_deep_field.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/english/Horse/SameSire.aspx?HorseSire=Deep%20Field", content=f.read())
            with open(os.path.join(self.dir, "jockey_YML.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Jockey/JockeyProfile.aspx?JockeyId=YML&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "trainer_HDA.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Trainers/TrainerProfile.aspx?TrainerId=HDA&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "horse_HK_2024_K429.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Horse/Horse.aspx?HorseId=HK_2024_K429", content=f.read())
            with open(os.path.join(self.dir, "sire_alabama_express.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/english/Horse/SameSire.aspx?HorseSire=Alabama%20Express", content=f.read())
            with open(os.path.join(self.dir, "jockey_PMF.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Jockey/JockeyProfile.aspx?JockeyId=PMF&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "horse_HK_2024_K464.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Horse/Horse.aspx?HorseId=HK_2024_K464", content=f.read())
            with open(os.path.join(self.dir, "sire_siyouni.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/english/Horse/SameSire.aspx?HorseSire=Siyouni", content=f.read())
            with open(os.path.join(self.dir, "jockey_BH.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Jockey/JockeyProfile.aspx?JockeyId=BH&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "trainer_FC.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Trainers/TrainerProfile.aspx?TrainerId=FC&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "horse_HK_2024_K331.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Horse/Horse.aspx?HorseId=HK_2024_K331", content=f.read())
            with open(os.path.join(self.dir, "sire_star_turn.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Horse/SameSire.aspx?HorseSire=Star%20Turn", content=f.read())
            with open(os.path.join(self.dir, "jockey_FEL.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Jockey/JockeyProfile.aspx?JockeyId=FEL&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "trainer_EDJ.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Trainers/TrainerProfile.aspx?TrainerId=EDJ&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "horse_HK_2024_K259.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Horse/Horse.aspx?HorseId=HK_2024_K259", content=f.read())
            with open(os.path.join(self.dir, "sire_oasis_dream.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/english/Horse/SameSire.aspx?HorseSire=Oasis%20Dream", content=f.read())
            with open(os.path.join(self.dir, "jockey_AA.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Jockey/JockeyProfile.aspx?JockeyId=AA&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "horse_HK_2024_K157.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Horse/Horse.aspx?HorseId=HK_2024_K157", content=f.read())
            with open(os.path.join(self.dir, "sire_nyquist.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/english/Horse/SameSire.aspx?HorseSire=Nyquist", content=f.read())
            with open(os.path.join(self.dir, "jockey_HEL.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Jockey/JockeyProfile.aspx?JockeyId=HEL&Season=Current", content=f.read())
            with open(os.path.join(self.dir, "trainer_LKW.html"), "rb") as f:
                m.get("https://racing.hkjc.com/racing/information/English/Trainers/TrainerProfile.aspx?TrainerId=LKW&Season=Current", content=f.read())
            game_model = create_hkjc_hkjc_game_model(
                session=self._session,
                html=html,
                url="https://racing.hkjc.com/racing/information/English/Racing/LocalResults.aspx?RaceDate=2025/05/25&Racecourse=ST&RaceNo=1",
            )
            self.assertEqual(game_model.teams[-1].players[0].weight, 486.250624)

    def test_empty_race(self):
        with requests_mock.Mocker() as m:
            html = ""
            with open(os.path.join(self.dir, "game2.html"), "r") as f:
                html = f.read()
            game_model = create_hkjc_hkjc_game_model(
                session=self._session,
                html=html,
                url="https://racing.hkjc.com/racing/information/English/Racing/LocalResults.aspx?RaceDate=2025/05/26&Racecourse=HV&RaceNo=1",
            )
            self.assertIsNone(game_model)

    def test_race_declared_abandoned(self):
        with requests_mock.Mocker() as m:
            html = ""
            with open(os.path.join(self.dir, "game3.html"), "r") as f:
                html = f.read()
            game_model = create_hkjc_hkjc_game_model(
                session=self._session,
                html=html,
                url="https://racing.hkjc.com/racing/information/English/Racing/LocalResults.aspx?RaceDate=2024%2F11%2F13&RaceNo=7&Racecourse=HV",
            )
            self.assertIsNone(game_model)



"""Aussportsbetting league model."""

# pylint: disable=line-too-long
from io import BytesIO
from typing import Any, Iterator

import tqdm
from dateutil.parser import parse
from openpyxl import load_workbook
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ..game_model import GameModel
from ..league import League
from ..league_model import SHUTDOWN_FLAG, LeagueModel, needs_shutdown
from .aussportsbetting_game_model import create_aussportsbetting_game_model


class AusSportsBettingLeagueModel(LeagueModel):
    """AusSportsBetting implementation of the league model."""

    def __init__(
        self,
        league: League,
        session: ScrapeSession,
        position: int | None = None,
    ) -> None:
        super().__init__(league, session, position=position)
        match league:
            case League.AFL:
                self._spreadsheet_url = (
                    "https://www.aussportsbetting.com/historical_data/afl.xlsx"
                )
            case League.NFL:
                self._spreadsheet_url = (
                    "https://www.aussportsbetting.com/historical_data/nfl.xlsx"
                )
            case _:
                raise ValueError(
                    f"League {league} not supported by aus sports betting."
                )

    @classmethod
    def name(cls) -> str:
        return "aussportsbetting-league-model"

    def _row_to_game(self, row: Any) -> GameModel | None:
        current_cell_idx = 0
        date_cell = str(row[0].value)
        if date_cell in {"Date", "None"}:
            return None
        current_cell_idx += 1
        if self.league == League.AFL:
            time_cell = str(row[current_cell_idx].value)
            dt = parse(" ".join([date_cell, time_cell]))
            current_cell_idx += 1
        else:
            dt = parse(date_cell)
        home_team = str(row[current_cell_idx].value).strip()
        current_cell_idx += 1
        away_team = str(row[current_cell_idx].value).strip()
        current_cell_idx += 1
        venue = None
        if self.league == League.AFL:
            venue = str(row[current_cell_idx].value).strip()
            current_cell_idx += 1
        home_points = float(row[current_cell_idx].value)  # type: ignore
        current_cell_idx += 1
        away_points = float(row[current_cell_idx].value)  # type: ignore
        current_cell_idx += 1

        if self.league == League.NFL:
            current_cell_idx += 1
        play_off = row[current_cell_idx].value == "Y"

        if self.league == League.NFL:
            current_cell_idx += 2
        else:
            current_cell_idx += 5
        home_odds = float(row[current_cell_idx].value)  # type: ignore
        current_cell_idx += 1
        if self.league == League.NFL:
            current_cell_idx += 3
        away_odds = float(row[current_cell_idx].value)  # type: ignore
        current_cell_idx += 1
        return create_aussportsbetting_game_model(
            dt,
            home_team,
            away_team,
            venue,
            self.session,
            home_points,
            away_points,
            home_odds,
            away_odds,
            self.league,
            play_off,
        )

    @property
    def games(self) -> Iterator[GameModel]:
        self.session.cache.delete(urls=[self._spreadsheet_url])
        try:
            with self.session.cache_disabled():
                response = self.session.get(self._spreadsheet_url)
            response.raise_for_status()
            workbook = load_workbook(filename=BytesIO(response.content))
            ws = workbook.active
            if ws is None:
                raise ValueError("ws is null.")
            with tqdm.tqdm(position=self.position) as pbar:
                for row in ws.iter_rows():
                    if needs_shutdown():
                        return
                    game_model = self._row_to_game(row)
                    if game_model is not None:
                        pbar.update(1)
                        pbar.set_description(f"AusSportsBetting - {game_model.dt}")
                        yield game_model
        except Exception as exc:
            SHUTDOWN_FLAG.set()
            raise exc



"""Tests for the NFL sportsdb league model class."""
import unittest

import requests_cache
from sportsball.data.nfl.sportsdb.nfl_sportsdb_league_model import NFLSportsDBLeagueModel
from sportsball.data.league import League


class TestNFLSportsDBLeagueModel(unittest.TestCase):

    def setUp(self):
        self.session = requests_cache.CachedSession(backend="memory")
        self.model = NFLSportsDBLeagueModel(self.session)

    def test_league(self):
        self.assertEqual(self.model.league, League.NFL)



"""MLB combined league model."""

from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ...combined.combined_league_model import CombinedLeagueModel
from ...league import League
from ..espn.mlb_espn_league_model import MLBESPNLeagueModel
from ..oddsportal.mlb_oddsportal_league_model import MLBOddsPortalLeagueModel

MLB_TEAM_IDENTITY_MAP: dict[str, str] = {}
MLB_VENUE_IDENTITY_MAP: dict[str, str] = {}
MLB_PLAYER_IDENTITY_MAP: dict[str, str] = {}


class MLBCombinedLeagueModel(CombinedLeagueModel):
    """MLB combined implementation of the league model."""

    def __init__(self, session: ScrapeSession, league_filter: str | None) -> None:
        super().__init__(
            session,
            League.MLB,
            [
                MLBESPNLeagueModel(session, position=0),
                MLBOddsPortalLeagueModel(session, position=1),
                # MLBSportsDBLeagueModel(session, position=2),
                # MLBSportsReferenceLeagueModel(session, position=3),
            ],
            league_filter,
        )

    @classmethod
    def name(cls) -> str:
        return "mlb-combined-league-model"

    @classmethod
    def team_identity_map(cls) -> dict[str, str]:
        return MLB_TEAM_IDENTITY_MAP

    @classmethod
    def venue_identity_map(cls) -> dict[str, str]:
        return MLB_VENUE_IDENTITY_MAP

    @classmethod
    def player_identity_map(cls) -> dict[str, str]:
        return MLB_PLAYER_IDENTITY_MAP



"""MLB ESPN league model."""

from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ...espn.espn_league_model import ESPNLeagueModel
from ...league import League
from ..position import Position

_SEASON_URL = (
    "https://sports.core.api.espn.com/v2/sports/baseball/leagues/mlb/seasons?limit=100"
)


class MLBESPNLeagueModel(ESPNLeagueModel):
    """MLB ESPN implementation of the league model."""

    def __init__(self, session: ScrapeSession, position: int | None = None) -> None:
        super().__init__(_SEASON_URL, League.MLB, session, position=position)

    @classmethod
    def name(cls) -> str:
        return "mlb-espn-league-model"

    @classmethod
    def position_validator(cls) -> dict[str, str]:
        return {str(x): str(x) for x in Position}



"""The CLI for executing the data harvesting."""

import datetime
import io
import logging
import sys
from contextlib import redirect_stdout

from . import __VERSION__
from .args import parse_args
from .data import league_model
from .data.league import league_from_str
from .logger import setup_logger
from .sportsball import SportsBall

_STDOUT_FILE = "-"


def main() -> None:
    """The main CLI function."""
    with redirect_stdout(sys.stderr):
        args = parse_args()
        setup_logger()

        logging.info("--- sportsball %s ---", __VERSION__)

        if args.timeout is not None:
            league_model.TIMEOUT_DT = datetime.datetime.now() + datetime.timedelta(
                seconds=args.timeout
            )

        ball = SportsBall()
        league = ball.league(league_from_str(args.league), args.leaguemodel)
        df = league.to_frame()
        handle = io.BytesIO()
        df.to_parquet(handle, compression="gzip")
        handle.seek(0)
    if args.file == _STDOUT_FILE:
        sys.stdout.buffer.write(handle.getbuffer())
    else:
        with open(args.file, "wb") as fhandle:
            fhandle.write(handle.getbuffer())


if __name__ == "__main__":
    main()



"""Combined weather model."""

# pylint: disable=too-many-locals,too-many-statements,too-many-branches
from ..weather_model import VERSION, WeatherModel
from .most_interesting import more_interesting


def create_combined_weather_model(
    weather_models: list[WeatherModel],
) -> WeatherModel | None:
    """Create a weather model by combining many weather models."""
    if not weather_models:
        return None
    temperature = None
    relative_humidity = None
    dew_point = None
    apparent_temperature = None
    precipitation_probability = None
    precipitation = None
    rain = None
    showers = None
    snowfall = None
    snow_depth = None
    weather_code = None
    sealevel_pressure = None
    surface_pressure = None
    cloud_cover_total = None
    cloud_cover_low = None
    cloud_cover_mid = None
    cloud_cover_high = None
    visibility = None
    evapotranspiration = None
    reference_evapotranspiration = None
    vapour_pressure_deficit = None
    wind_speed_10m = None
    wind_speed_80m = None
    wind_speed_120m = None
    wind_speed_180m = None
    wind_direction_10m = None
    wind_direction_80m = None
    wind_direction_120m = None
    wind_direction_180m = None
    wind_gusts = None
    temperature_80m = None
    temperature_120m = None
    temperature_180m = None
    soil_temperature_0cm = None
    soil_temperature_6cm = None
    soil_temperature_18cm = None
    soil_temperature_54cm = None
    soil_moisture_0cm = None
    soil_moisture_1cm = None
    soil_moisture_3cm = None
    soil_moisture_9cm = None
    soil_moisture_27cm = None
    daily_weather_code = None
    daily_maximum_temperature_2m = None
    daily_minimum_temperature_2m = None
    daily_maximum_apparent_temperature_2m = None
    daily_minimum_apparent_temperature_2m = None
    sunrise = None
    sunset = None
    daylight_duration = None
    sunshine_duration = None
    uv_index = None
    uv_index_clear_sky = None
    rain_sum = None
    showers_sum = None
    snowfall_sum = None
    precipitation_sum = None
    precipitation_hours = None
    precipitation_probability_max = None
    maximum_wind_speed_10m = None
    maximum_wind_gusts_10m = None
    dominant_wind_direction = None
    shortwave_radiation_sum = None
    daily_reference_evapotranspiration = None
    for weather_model in weather_models:
        temperature = more_interesting(temperature, weather_model.temperature)
        relative_humidity = more_interesting(
            relative_humidity, weather_model.relative_humidity
        )
        dew_point = more_interesting(dew_point, weather_model.dew_point)
        apparent_temperature = more_interesting(
            apparent_temperature, weather_model.apparent_temperature
        )
        precipitation_probability = more_interesting(
            precipitation_probability, weather_model.precipitation_probability
        )
        precipitation = more_interesting(precipitation, weather_model.precipitation)
        rain = more_interesting(rain, weather_model.rain)
        showers = more_interesting(showers, weather_model.showers)
        snowfall = more_interesting(snowfall, weather_model.snowfall)
        snow_depth = more_interesting(snow_depth, weather_model.snow_depth)
        weather_code = more_interesting(weather_code, weather_model.weather_code)
        sealevel_pressure = more_interesting(
            sealevel_pressure, weather_model.sealevel_pressure
        )
        surface_pressure = more_interesting(
            surface_pressure, weather_model.surface_pressure
        )
        cloud_cover_total = more_interesting(
            cloud_cover_total, weather_model.cloud_cover_total
        )
        cloud_cover_low = more_interesting(
            cloud_cover_low, weather_model.cloud_cover_low
        )
        cloud_cover_mid = more_interesting(
            cloud_cover_mid, weather_model.cloud_cover_mid
        )
        cloud_cover_high = more_interesting(
            cloud_cover_high, weather_model.cloud_cover_high
        )
        visibility = more_interesting(visibility, weather_model.visibility)
        evapotranspiration = more_interesting(
            evapotranspiration, weather_model.evapotranspiration
        )
        reference_evapotranspiration = more_interesting(
            reference_evapotranspiration, weather_model.reference_evapotranspiration
        )
        vapour_pressure_deficit = more_interesting(
            vapour_pressure_deficit, weather_model.vapour_pressure_deficit
        )
        wind_speed_10m = more_interesting(wind_speed_10m, weather_model.wind_speed_10m)
        wind_speed_80m = more_interesting(wind_speed_80m, weather_model.wind_speed_80m)
        wind_speed_120m = more_interesting(
            wind_speed_120m, weather_model.wind_speed_120m
        )
        wind_speed_180m = more_interesting(
            wind_speed_180m, weather_model.wind_speed_180m
        )
        wind_direction_10m = more_interesting(
            wind_direction_10m, weather_model.wind_direction_10m
        )
        wind_direction_80m = more_interesting(
            wind_direction_80m, weather_model.wind_direction_80m
        )
        wind_direction_120m = more_interesting(
            wind_direction_120m, weather_model.wind_direction_120m
        )
        wind_direction_180m = more_interesting(
            wind_direction_180m, weather_model.wind_direction_180m
        )
        wind_gusts = more_interesting(wind_gusts, weather_model.wind_gusts)
        temperature_80m = more_interesting(
            temperature_80m, weather_model.temperature_80m
        )
        temperature_120m = more_interesting(
            temperature_120m, weather_model.temperature_120m
        )
        temperature_180m = more_interesting(
            temperature_180m, weather_model.temperature_180m
        )
        soil_temperature_0cm = more_interesting(
            soil_temperature_0cm, weather_model.soil_temperature_0cm
        )
        soil_temperature_6cm = more_interesting(
            soil_temperature_6cm, weather_model.soil_temperature_6cm
        )
        soil_temperature_18cm = more_interesting(
            soil_temperature_18cm, weather_model.soil_temperature_18cm
        )
        soil_temperature_54cm = more_interesting(
            soil_temperature_54cm, weather_model.soil_temperature_54cm
        )
        soil_moisture_0cm = more_interesting(
            soil_moisture_0cm, weather_model.soil_moisture_0cm
        )
        soil_moisture_1cm = more_interesting(
            soil_moisture_1cm, weather_model.soil_moisture_1cm
        )
        soil_moisture_3cm = more_interesting(
            soil_moisture_3cm, weather_model.soil_moisture_3cm
        )
        soil_moisture_9cm = more_interesting(
            soil_moisture_9cm, weather_model.soil_moisture_9cm
        )
        soil_moisture_27cm = more_interesting(
            soil_moisture_27cm, weather_model.soil_moisture_27cm
        )
        daily_weather_code = more_interesting(
            daily_weather_code, weather_model.daily_weather_code
        )
        daily_maximum_temperature_2m = more_interesting(
            daily_maximum_temperature_2m, weather_model.daily_maximum_temperature_2m
        )
        daily_minimum_temperature_2m = more_interesting(
            daily_minimum_temperature_2m, weather_model.daily_minimum_temperature_2m
        )
        daily_maximum_apparent_temperature_2m = more_interesting(
            daily_maximum_apparent_temperature_2m,
            weather_model.daily_maximum_apparent_temperature_2m,
        )
        daily_minimum_apparent_temperature_2m = more_interesting(
            daily_minimum_apparent_temperature_2m,
            weather_model.daily_minimum_apparent_temperature_2m,
        )
        sunrise = more_interesting(sunrise, weather_model.sunrise)
        sunset = more_interesting(sunset, weather_model.sunset)
        daylight_duration = more_interesting(
            daylight_duration, weather_model.daylight_duration
        )
        sunshine_duration = more_interesting(
            sunshine_duration, weather_model.sunshine_duration
        )
        uv_index = more_interesting(uv_index, weather_model.uv_index)
        uv_index_clear_sky = more_interesting(
            uv_index_clear_sky, weather_model.uv_index_clear_sky
        )
        rain_sum = more_interesting(rain_sum, weather_model.rain_sum)
        showers_sum = more_interesting(showers_sum, weather_model.showers_sum)
        snowfall_sum = more_interesting(snowfall_sum, weather_model.snowfall_sum)
        precipitation_sum = more_interesting(
            precipitation_sum, weather_model.precipitation_sum
        )
        precipitation_hours = more_interesting(
            precipitation_hours, weather_model.precipitation_hours
        )
        precipitation_probability_max = more_interesting(
            precipitation_probability_max, weather_model.precipitation_probability_max
        )
        maximum_wind_speed_10m = more_interesting(
            maximum_wind_speed_10m, weather_model.maximum_wind_speed_10m
        )
        maximum_wind_gusts_10m = more_interesting(
            maximum_wind_gusts_10m, weather_model.maximum_wind_gusts_10m
        )
        dominant_wind_direction = more_interesting(
            dominant_wind_direction, weather_model.dominant_wind_direction
        )
        shortwave_radiation_sum = more_interesting(
            shortwave_radiation_sum, weather_model.shortwave_radiation_sum
        )
        daily_reference_evapotranspiration = more_interesting(
            daily_reference_evapotranspiration,
            weather_model.daily_reference_evapotranspiration,
        )

    return WeatherModel.model_construct(
        temperature=temperature,
        relative_humidity=relative_humidity,
        dew_point=dew_point,
        apparent_temperature=apparent_temperature,
        precipitation_probability=precipitation_probability,
        precipitation=precipitation,
        rain=rain,
        showers=showers,
        snowfall=snowfall,
        snow_depth=snow_depth,
        weather_code=weather_code,
        sealevel_pressure=sealevel_pressure,
        surface_pressure=surface_pressure,
        cloud_cover_total=cloud_cover_total,
        cloud_cover_low=cloud_cover_low,
        cloud_cover_mid=cloud_cover_mid,
        cloud_cover_high=cloud_cover_high,
        visibility=visibility,
        evapotranspiration=evapotranspiration,
        reference_evapotranspiration=reference_evapotranspiration,
        vapour_pressure_deficit=vapour_pressure_deficit,
        wind_speed_10m=wind_speed_10m,
        wind_speed_80m=wind_speed_80m,
        wind_speed_120m=wind_speed_120m,
        wind_speed_180m=wind_speed_180m,
        wind_direction_10m=wind_direction_10m,
        wind_direction_80m=wind_direction_80m,
        wind_direction_120m=wind_direction_120m,
        wind_direction_180m=wind_direction_180m,
        wind_gusts=wind_gusts,
        temperature_80m=temperature_80m,
        temperature_120m=temperature_120m,
        temperature_180m=temperature_180m,
        soil_temperature_0cm=soil_temperature_0cm,
        soil_temperature_6cm=soil_temperature_6cm,
        soil_temperature_18cm=soil_temperature_18cm,
        soil_temperature_54cm=soil_temperature_54cm,
        soil_moisture_0cm=soil_moisture_0cm,
        soil_moisture_1cm=soil_moisture_1cm,
        soil_moisture_3cm=soil_moisture_3cm,
        soil_moisture_9cm=soil_moisture_9cm,
        soil_moisture_27cm=soil_moisture_27cm,
        daily_weather_code=daily_weather_code,
        daily_maximum_temperature_2m=daily_maximum_temperature_2m,
        daily_minimum_temperature_2m=daily_minimum_temperature_2m,
        daily_maximum_apparent_temperature_2m=daily_maximum_apparent_temperature_2m,
        daily_minimum_apparent_temperature_2m=daily_minimum_apparent_temperature_2m,
        sunrise=sunrise,
        sunset=sunset,
        daylight_duration=daylight_duration,
        sunshine_duration=sunshine_duration,
        uv_index=uv_index,
        uv_index_clear_sky=uv_index_clear_sky,
        rain_sum=rain_sum,
        showers_sum=showers_sum,
        snowfall_sum=snowfall_sum,
        precipitation_sum=precipitation_sum,
        precipitation_hours=precipitation_hours,
        precipitation_probability_max=precipitation_probability_max,
        maximum_wind_speed_10m=maximum_wind_speed_10m,
        maximum_wind_gusts_10m=maximum_wind_gusts_10m,
        dominant_wind_direction=dominant_wind_direction,
        shortwave_radiation_sum=shortwave_radiation_sum,
        daily_reference_evapotranspiration=daily_reference_evapotranspiration,
        version=VERSION,
    )



"""The prototype class for a venue."""

from typing import Literal

from pydantic import BaseModel, ConfigDict, Field

from .address_model import VERSION as ADDRESS_VERSION
from .address_model import AddressModel
from .delimiter import DELIMITER
from .field_type import TYPE_KEY, FieldType

VENUE_IDENTIFIER_COLUMN: Literal["identifier"] = "identifier"
VENUE_ADDRESS_COLUMN: Literal["address"] = "address"
VENUE_IS_TURF_COLUMN: Literal["is_turf"] = "is_turf"
VENUE_IS_DIRT_COLUMN: Literal["is_dirt"] = "is_dirt"
VERSION = DELIMITER.join(["0.0.1", ADDRESS_VERSION])


class VenueModel(BaseModel):
    """The serialisable venue class."""

    model_config = ConfigDict(
        validate_assignment=False,
        revalidate_instances="never",
        extra="ignore",
        from_attributes=False,
    )

    identifier: str = Field(
        ...,
        json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL},
        alias=VENUE_IDENTIFIER_COLUMN,
    )
    name: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.TEXT})
    address: AddressModel | None = Field(..., alias=VENUE_ADDRESS_COLUMN)
    is_grass: bool | None
    is_indoor: bool | None
    is_turf: bool | None = Field(..., alias=VENUE_IS_TURF_COLUMN)
    is_dirt: bool | None = Field(..., alias=VENUE_IS_DIRT_COLUMN)
    version: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL})
